# -*- coding: utf-8 -*-

from hbshare.fe.xwq.analysis.orm.hbdb import HBDB
from datetime import datetime, timedelta
import os
import numpy as np
import pandas as pd


def get_date(start_date, end_date):
    calendar_df = HBDB().read_cal(start_date, end_date)
    calendar_df = calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
    calendar_df['CALENDAR_DATE'] = calendar_df['CALENDAR_DATE'].astype(str)
    calendar_df = calendar_df.sort_values('CALENDAR_DATE')
    calendar_df['IS_OPEN'] = calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
    calendar_df['YEAR_MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[:6])
    calendar_df['MONTH'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:6])
    calendar_df['MONTH_DAY'] = calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:])
    calendar_df = calendar_df[(calendar_df['CALENDAR_DATE'] >= start_date) & (calendar_df['CALENDAR_DATE'] <= end_date)]
    trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    trade_df = trade_df[(trade_df['TRADE_DATE'] >= start_date) & (trade_df['TRADE_DATE'] <= end_date)]
    report_df = calendar_df.drop_duplicates('YEAR_MONTH', keep='last').rename(columns={'CALENDAR_DATE': 'REPORT_DATE'})
    report_df = report_df[report_df['MONTH_DAY'].isin(['0331', '0630', '0930', '1231'])]
    report_df = report_df[(report_df['REPORT_DATE'] >= start_date) & (report_df['REPORT_DATE'] <= end_date)]
    report_trade_df = calendar_df[calendar_df['IS_OPEN'] == 1].rename(columns={'CALENDAR_DATE': 'TRADE_DATE'})
    report_trade_df = report_trade_df.sort_values('TRADE_DATE').drop_duplicates('YEAR_MONTH', keep='last')
    report_trade_df = report_trade_df[report_trade_df['MONTH'].isin(['03', '06', '09', '12'])]
    report_trade_df = report_trade_df[(report_trade_df['TRADE_DATE'] >= start_date) & (report_trade_df['TRADE_DATE'] <= end_date)]
    calendar_trade_df = calendar_df[['CALENDAR_DATE']].merge(trade_df[['TRADE_DATE']], left_on=['CALENDAR_DATE'], right_on=['TRADE_DATE'], how='left')
    calendar_trade_df['TRADE_DATE'] = calendar_trade_df['TRADE_DATE'].fillna(method='ffill')
    calendar_trade_df = calendar_trade_df[(calendar_trade_df['TRADE_DATE'] >= start_date) & (calendar_trade_df['TRADE_DATE'] <= end_date)]
    return calendar_df, report_df, trade_df, report_trade_df, calendar_trade_df

def get_stock_industry():
    # stock_industry = HBDB().read_stock_industry()
    # stock_industry.to_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/industry_tracking/stock_industry.hdf', key='table', mode='w')
    stock_industry = pd.read_hdf('D:/Git/hbshare/hbshare/fe/xwq/data/industry_tracking/stock_industry.hdf', key='table')
    stock_industry = stock_industry.rename(columns={'zqdm': 'TICKER_SYMBOL', 'flmc': 'INDUSTRY_NAME', 'fldm': 'INDUSTRY_ID', 'hyhfbz': 'INDUSTRY_VERSION', 'fljb': 'INDUSTRY_TYPE', 'qsrq': 'BEGIN_DATE', 'jsrq': 'END_DATE', 'sfyx': 'IS_NEW'})
    stock_industry = stock_industry.dropna(subset=['BEGIN_DATE'])
    stock_industry['END_DATE'] = stock_industry['END_DATE'].fillna('20990101')
    stock_industry['BEGIN_DATE'] = stock_industry['BEGIN_DATE'].astype(int).astype(str)
    stock_industry['END_DATE'] = stock_industry['END_DATE'].astype(int).astype(str)
    stock_industry['INDUSTRY_VERSION'] = stock_industry['INDUSTRY_VERSION'].astype(int)
    stock_industry['INDUSTRY_TYPE'] = stock_industry['INDUSTRY_TYPE'].astype(int)
    stock_industry['IS_NEW'] = stock_industry['IS_NEW'].astype(int)
    stock_industry = stock_industry[stock_industry['INDUSTRY_VERSION'] == 2]
    stock_industry = stock_industry.loc[stock_industry['TICKER_SYMBOL'].str.len() == 6]
    stock_industry = stock_industry.loc[stock_industry['TICKER_SYMBOL'].astype(str).str.slice(0, 1).isin(['0', '3', '6'])]
    return stock_industry

def get_industry_info():
    industry_info = HBDB().read_industry_info()
    industry_info = industry_info.rename(columns={'flmc': 'INDUSTRY_NAME', 'zsdm': 'INDUSTRY_ID', 'hyhfbz': 'INDUSTRY_VERSION', 'fljb': 'INDUSTRY_TYPE', 'qsrq': 'BEGIN_DATE', 'jsrq': 'END_DATE', 'sfyx': 'IS_NEW'})
    industry_info = industry_info.dropna(subset=['BEGIN_DATE'])
    industry_info['END_DATE'] = industry_info['END_DATE'].replace('', np.nan).fillna('20990101')
    industry_info['BEGIN_DATE'] = industry_info['BEGIN_DATE'].apply(lambda x: datetime.strptime(x, '%Y/%m/%d').strftime('%Y%m%d') if '/' in x else x)
    industry_info['END_DATE'] = industry_info['END_DATE'].apply(lambda x: datetime.strptime(x, '%Y/%m/%d').strftime('%Y%m%d') if '/' in x else x)
    industry_info['BEGIN_DATE'] = industry_info['BEGIN_DATE'].astype(int).astype(str)
    industry_info['END_DATE'] = industry_info['END_DATE'].astype(int).astype(str)
    industry_info['INDUSTRY_VERSION'] = industry_info['INDUSTRY_VERSION'].astype(int)
    industry_info['INDUSTRY_TYPE'] = industry_info['INDUSTRY_TYPE'].astype(int)
    industry_info['IS_NEW'] = industry_info['IS_NEW'].astype(int)
    industry_info = industry_info[industry_info['INDUSTRY_VERSION'] == 3]
    return industry_info

def get_stock_info():
    stock_info = HBDB().read_stock_info()
    stock_info = stock_info.rename(columns={'zqdm': 'TICKER_SYMBOL', 'zqjc': 'SEC_SHORT_NAME', 'ssrq': 'ESTABLISH_DATE'})
    stock_info['ESTABLISH_DATE'] = stock_info['ESTABLISH_DATE'].dropna()
    stock_info['ESTABLISH_DATE'] = stock_info['ESTABLISH_DATE'].astype(int).astype(str)
    stock_info = stock_info.loc[stock_info['TICKER_SYMBOL'].str.len() == 6]
    stock_info = stock_info.loc[stock_info['TICKER_SYMBOL'].astype(str).str.slice(0, 1).isin(['0', '3', '6'])]
    stock_info['SAMPLE_DATE'] = stock_info['ESTABLISH_DATE'].apply(lambda x: (datetime.strptime(x, '%Y%m%d') + timedelta(365)).strftime('%Y%m%d'))
    return stock_info

class IndustryHead:
    def __init__(self, date, start_date, end_date, sw_type, head):
        self.date = date
        self.start_date = start_date
        self.end_date = end_date
        self.sw_type = sw_type
        self.head = head
        self.load_data()

    def load_data(self):
        self.calendar_df, self.report_df, self.trade_df, self.report_trade_df, self.calendar_trade_df = get_date(self.start_date, self.end_date)

        self.industry_info = get_industry_info()
        self.industry_info = self.industry_info[self.industry_info['INDUSTRY_TYPE'] == self.sw_type]
        self.industry_info = self.industry_info[self.industry_info['IS_NEW'] == 1]
        self.industry_info = self.industry_info[['INDUSTRY_NAME', 'INDUSTRY_ID', 'BEGIN_DATE', 'END_DATE']]
        self.industry_id_name_dic = self.industry_info[['INDUSTRY_ID', 'INDUSTRY_NAME']].set_index('INDUSTRY_ID')['INDUSTRY_NAME'].to_dict()
        self.industry_name_id_dic = self.industry_info[['INDUSTRY_ID', 'INDUSTRY_NAME']].set_index('INDUSTRY_NAME')['INDUSTRY_ID'].to_dict()

        self.stock_industry = get_stock_industry()
        self.stock_industry = self.stock_industry[self.stock_industry['INDUSTRY_TYPE'] == self.sw_type]
        self.stock_industry = self.stock_industry[self.stock_industry['IS_NEW'] == 1]
        self.stock_industry = self.stock_industry[['INDUSTRY_NAME', 'TICKER_SYMBOL', 'BEGIN_DATE', 'END_DATE']]

        stock_market_value = HBDB().read_stock_market_value_given_date(self.date)
        star_stock_market_value = HBDB().read_star_stock_market_value_given_date(self.date)
        self. stock_market_value = pd.concat([stock_market_value, star_stock_market_value])

        stock_finance_path = 'D:/Git/hbshare/hbshare/fe/xwq/data/industry_tracking/stock_finance.hdf'
        # if os.path.isfile(stock_finance_path):
        #     existed_stock_finance = pd.read_hdf(stock_finance_path, key='table')
        #     max_date = max(existed_stock_finance['END_DATE'])
        #     start_date = max(str(max_date), '20150101')
        # else:
        #     existed_stock_finance = pd.DataFrame()
        #     start_date = '20150101'
        # report_df = self.report_df[(self.report_df['REPORT_DATE'] > start_date) & (self.report_df['REPORT_DATE'] < datetime.today().strftime('%Y%m%d'))]
        # stock_finance_list = []
        # for date in report_df['REPORT_DATE'].unique().tolist():
        #     stock_finance_date = HBDB().read_stock_finance_given_date(date)
        #     stock_finance_date = stock_finance_date[['END_DATE', 'PUBLISH_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'NET_PROFIT', 'MAIN_INCOME_PS', 'ROE_TTM', 'GROSS_INCOME_RATIO_TTM', 'NET_PROFIT_RATIO_TTM', 'EPS_TTM', 'OPER_CASH_FLOW_PS_TTM', 'NET_ASSET_PS']] if len(stock_finance_date) > 0 else pd.DataFrame(columns=['END_DATE', 'PUBLISH_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'NET_PROFIT', 'MAIN_INCOME_PS', 'ROE_TTM', 'GROSS_INCOME_RATIO_TTM', 'NET_PROFIT_RATIO_TTM', 'EPS_TTM', 'OPER_CASH_FLOW_PS_TTM', 'NET_ASSET_PS'])
        #     star_stock_finance_date = HBDB().read_star_stock_finance_given_date(date)
        #     star_stock_finance_date = star_stock_finance_date[['END_DATE', 'PUBLISH_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'NET_PROFIT', 'MAIN_INCOME_PS', 'ROE_TTM', 'GROSS_INCOME_RATIO_TTM', 'NET_PROFIT_RATIO_TTM', 'EPS_TTM', 'OPER_CASH_FLOW_PS_TTM', 'NET_ASSET_PS']] if len(star_stock_finance_date) > 0 else pd.DataFrame(columns=['END_DATE', 'PUBLISH_DATE', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'NET_PROFIT', 'MAIN_INCOME_PS', 'ROE_TTM', 'GROSS_INCOME_RATIO_TTM', 'NET_PROFIT_RATIO_TTM', 'EPS_TTM', 'OPER_CASH_FLOW_PS_TTM', 'NET_ASSET_PS'])
        #     stock_finance_date = pd.concat([stock_finance_date, star_stock_finance_date])
        #     stock_finance_list.append(stock_finance_date)
        #     print(date)
        # self.stock_finance = pd.concat([existed_stock_finance] + stock_finance_list, ignore_index=True)
        # self.stock_finance = self.stock_finance.sort_values(['END_DATE', 'TICKER_SYMBOL', 'PUBLISH_DATE'])
        # self.stock_finance = self.stock_finance.reset_index().drop('index', axis=1)
        # self.stock_finance.to_hdf(stock_finance_path, key='table', mode='w')
        self.stock_finance = pd.read_hdf(stock_finance_path, key='table')

        stock_consensus_list = []
        for type in ['FY0', 'FY1', 'FY2', 'FY3', 'YOY']:
            stock_consensus_type = HBDB().read_consensus_given_date(self.date, type)
            stock_consensus_list.append(stock_consensus_type)
        self.stock_consensus = pd.concat(stock_consensus_list)
        self.stock_consensus['TICKER_SYMBOL'] = self.stock_consensus['TICKER_SYMBOL'].apply(lambda x: x.split('.')[0])

    def industry_head_stocks(self):
        stock_market_value = self.stock_market_value.merge(self.stock_industry[['TICKER_SYMBOL', 'INDUSTRY_NAME']], on=['TICKER_SYMBOL'], how='inner')
        stock_market_value = stock_market_value.dropna(subset=['TICKER_SYMBOL', 'INDUSTRY_NAME', 'MARKET_VALUE'])
        industry_head_stocks = stock_market_value.sort_values(['INDUSTRY_NAME', 'MARKET_VALUE'], ascending=[True, False]).groupby(['INDUSTRY_NAME']).head(self.head)
        industry_head_stocks = industry_head_stocks[['INDUSTRY_NAME', 'TICKER_SYMBOL', 'SEC_SHORT_NAME', 'MARKET_VALUE']]
        industry_head_stocks = industry_head_stocks.sort_values(['INDUSTRY_NAME', 'MARKET_VALUE'], ascending=[True, False])
        industry_head_stocks['MARKET_VALUE'] = industry_head_stocks['MARKET_VALUE'].apply(lambda x: round(x / 10000000000.0, 2))
        return industry_head_stocks

    def get_all(self):
        industry_head_stocks = self.industry_head_stocks()
        stock_consensus = self.stock_consensus[self.stock_consensus['TICKER_SYMBOL'].isin(industry_head_stocks['TICKER_SYMBOL'].unique().tolist())]
        est_net_profit = stock_consensus[stock_consensus['ROLLING_TYPE'] != 'YOY']
        est_net_profit['EST_NET_PROFIT'] = est_net_profit['EST_NET_PROFIT'].apply(lambda x: '{0}'.format(round(x / 10000000000.0, 2)))
        est_net_profit = est_net_profit.pivot(index='TICKER_SYMBOL', columns='ROLLING_TYPE', values='EST_NET_PROFIT').reset_index()
        industry_head_stocks = industry_head_stocks.merge(est_net_profit, on=['TICKER_SYMBOL'], how='left')
        stock_finance = self.stock_finance.sort_values(['TICKER_SYMBOL', 'END_DATE', 'PUBLISH_DATE']).drop_duplicates(['TICKER_SYMBOL', 'END_DATE'], keep='last')
        stock_finance = stock_finance[stock_finance['TICKER_SYMBOL'].isin(industry_head_stocks['TICKER_SYMBOL'].unique().tolist())]
        stock_finance = stock_finance[stock_finance['END_DATE'].str.slice(4, 8) == '1231']
        stock_finance['YEAR'] = stock_finance['END_DATE'].apply(lambda x: x[:4])
        stock_finance = stock_finance.sort_values(['TICKER_SYMBOL', 'YEAR'], ascending=[True, True]).groupby(['TICKER_SYMBOL']).tail(2)
        stock_finance['ROE_TTM'] = stock_finance['ROE_TTM'].apply(lambda x: '{0}%'.format(round(x, 2)))
        stock_finance['GROSS_INCOME_RATIO_TTM'] = stock_finance['GROSS_INCOME_RATIO_TTM'].apply(lambda x: '{0}%'.format(round(x, 2)))
        stock_finance['NET_PROFIT_RATIO_TTM'] = stock_finance['NET_PROFIT_RATIO_TTM'].apply(lambda x: '{0}%'.format(round(x, 2)))
        roe = stock_finance.pivot(index='TICKER_SYMBOL', columns='YEAR', values='ROE_TTM').reset_index()
        gir = stock_finance.pivot(index='TICKER_SYMBOL', columns='YEAR', values='GROSS_INCOME_RATIO_TTM').reset_index()
        npr = stock_finance.pivot(index='TICKER_SYMBOL', columns='YEAR', values='NET_PROFIT_RATIO_TTM').reset_index()
        roe.columns = ['TICKER_SYMBOL'] + ['ROE_' + str(col) for col in list(roe.columns)[1:] if col != 'TICKER_SYMBOL']
        gir.columns = ['TICKER_SYMBOL'] + ['毛利率_' + str(col) for col in list(gir.columns)[1:] if col != 'TICKER_SYMBOL']
        npr.columns = ['TICKER_SYMBOL'] + ['净利率_' + str(col) for col in list(npr.columns)[1:] if col != 'TICKER_SYMBOL']
        industry_head_stocks = industry_head_stocks.merge(roe, on=['TICKER_SYMBOL'], how='left').merge(gir, on=['TICKER_SYMBOL'], how='left').merge(npr, on=['TICKER_SYMBOL'], how='left')
        return industry_head_stocks

if __name__ == '__main__':
    date = '20220715'  # 每周最后一个交易日
    start_date = (datetime.strptime(date, '%Y%m%d') - timedelta(730)).strftime('%Y%m%d')
    end_date = date
    sw_type = 1  # 申万行业层级
    head = 5  # 龙头个股数量
    IndustryHead(date, start_date, end_date, sw_type, head).get_all()


"""
市场结构的测试代码部分
"""
import pandas as pd
import numpy as np
import os
import statsmodels.api as sm
from hbshare.fe.common.util.data_loader import get_trading_day_list
from hbshare.quant.Kevin.quant_room.MarketStructure import MarketHist, get_market_turnover, get_style_factor, \
    AlphaSeries, MarketStructure
from tqdm import tqdm


data_path = "D:\\研究基地\\Analysis"


def run_daily_plot(start_date, end_date):
    date_list = get_trading_day_list(start_date, end_date)
    for date in tqdm(date_list):
        MarketHist(date, '000905').daily_plot()


def run_market_structure(start_date, end_date):
    date_list = get_trading_day_list(start_date, end_date)
    all_res = []
    for date in tqdm(date_list):
        res = MarketStructure(date, '000905').get_construct_result()
        all_res.append(res)

    structure_all = pd.concat(all_res)
    structure_all.to_csv(os.path.join(data_path, "market_structure.csv"))


def run_factor_test(start_date, end_date, mode="1"):
    if mode == "1":
        factor_df = get_market_turnover(start_date, end_date)
    elif mode == "2":
        factor_df = pd.read_csv('D:\\市场微观结构图\\market_factor_save\\win_ratio.csv', index_col=0)
        factor_df['trade_date'] = factor_df.index
        factor_df['trade_date'] = factor_df['trade_date'].apply(lambda x: str(x))
        factor_df = factor_df.set_index('trade_date')
    elif mode == "3":
        factor_df = get_style_factor(start_date, end_date)
        # vol_df = factor_df.rolling(5).std().loc[start_date:]
        # vol_df.rename(columns=lambda x: str(x) + "_vol", inplace=True)
        factor_df = factor_df.loc[start_date:][['size', 'btop', 'growth']]
        # factor_df = pd.concat([factor_df, vol_df], axis=1)
    else:
        factor_df = pd.read_csv('D:\\市场微观结构图\\market_factor_save\\ind_cr.csv', index_col=0)
        factor_df['trade_date'] = factor_df.index
        factor_df['trade_date'] = factor_df['trade_date'].apply(lambda x: str(x))
        factor_df = factor_df.set_index('trade_date')

    alpha_series = AlphaSeries(start_date, end_date).calculate()
    idx = factor_df.index.intersection(alpha_series.index)
    factor_df = factor_df.reindex(idx)
    alpha_excess = alpha_series.reindex(idx)
    factor_df['alpha_excess'] = alpha_excess

    corr_df = factor_df.apply(lambda x: x.rolling(250).corr(factor_df['alpha_excess']).dropna(), axis=0)

    return factor_df


def run_analysis(start_date, end_date):
    cs_vol_df = pd.read_csv('D:\\市场微观结构图\\market_factor_save\\cs_vol.csv', index_col=0)
    cs_vol_df['trade_date'] = cs_vol_df.index
    cs_vol_df['trade_date'] = cs_vol_df['trade_date'].map(str)
    cs_vol_df = cs_vol_df.set_index('trade_date')

    win_ratio_df = pd.read_csv('D:\\市场微观结构图\\market_factor_save\\win_ratio.csv', index_col=0)
    win_ratio_df['trade_date'] = win_ratio_df.index
    win_ratio_df['trade_date'] = win_ratio_df['trade_date'].apply(lambda x: str(x))
    win_ratio_df = win_ratio_df.set_index('trade_date')

    style_df = get_style_factor(start_date, end_date)
    style_df = style_df.loc[start_date:][['size', 'btop', 'growth']]

    ind_df = pd.read_csv('D:\\市场微观结构图\\market_factor_save\\ind_cr.csv', index_col=0)
    ind_df['trade_date'] = ind_df.index
    ind_df['trade_date'] = ind_df['trade_date'].apply(lambda x: str(x))
    ind_df = ind_df.set_index('trade_date')

    factor_df = cs_vol_df[['skew_300', 'skew_800', 'skew_1800']].merge(
        win_ratio_df[['win_ratio_all', 'lose_spread_all']], left_index=True, right_index=True).merge(
        style_df[['size']], left_index=True, right_index=True).merge(
        ind_df[['win_ratio_abs', 'win_ratio_relative']], left_index=True, right_index=True)

    alpha_excess = AlphaSeries(start_date, end_date).calculate()
    idx = factor_df.index.intersection(alpha_excess.index)
    factor_df = factor_df.reindex(idx)
    alpha_excess = alpha_excess.reindex(idx)
    factor_df['alpha_excess'] = alpha_excess


if __name__ == '__main__':
    # run_daily_plot('20211215', '20211215')
    # df = run_factor_test('20180101', '20211112', mode="4")
    # print(df)
    # run_market_structure('20200101', '20211108')
    run_analysis('20180101', '20211112')


"""
可转债估值测试
"""
import hbshare as hbs
import pandas as pd
import numpy as np
import re
from tqdm import tqdm
import pickle
import os
from datetime import datetime, timedelta
from cb_pricing_BS import cbPricingBlackScholes
from cb_pricing_Tree import cbPricingTree
from cb_pricing_MC import cbPricingMC
from WindPy import w

w.start()


class CbValuation:
    """
    计算对应正股的波动率和股票价格
    """
    def __init__(self, date, stock_ticker):
        self.date = date
        self.stock_ticker = stock_ticker
        self._load_data()

    def _load_data(self):
        pre_dt = datetime.strptime(self.date, '%Y/%m/%d') - timedelta(days=400)
        pre_date = pre_dt.strftime('%Y%m%d')
        end_date = datetime.strptime(self.date, '%Y/%m/%d').strftime('%Y%m%d')
        sql_script = "SELECT SYMBOL, TDATE, TCLOSE FROM finchina.CHDQUOTE WHERE" \
                     " TDATE >= {} and TDATE <= {} and SYMBOL = {}".format(pre_date, end_date, self.stock_ticker)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        data = pd.DataFrame(res['data']).rename(
            columns={"SYMBOL": "ticker", "TDATE": "trade_date", "TCLOSE": "price"})
        data['trade_date'] = data['trade_date'].apply(lambda x: str(x))
        data = data[data['price'] > 0]

        self.vol = data['price'].pct_change()[-250:].std() * np.sqrt(250)
        self.stock_price = data.set_index('trade_date').iloc[-1]['price']


class CbValuationTest:
    """
    批量计算可转债价值
    """
    def __init__(self, basic_info, bond_ytm):
        self.basic_info = basic_info
        self.bond_ytm = bond_ytm
        self.prepared_data = None
        self._load_data()
        self._load_cb_price()

    def _load_data(self):
        if os.path.exists("C:\\Users\\kai.zhang\\Desktop\\可转债\\cb_params.pkl"):
            with open("C:\\Users\\kai.zhang\\Desktop\\可转债\\cb_params.pkl", "rb") as f:
                self.prepared_data = pickle.load(f)
        else:
            basic_info = self.basic_info.copy()
            basic_info = basic_info[basic_info['发行信用等级'] != 'A']
            cb_id_list = basic_info['证券代码'].tolist()
            basic_info = basic_info.set_index('证券代码')
            prepared_data = {}
            for cb_id in tqdm(cb_id_list):
                ConvPrice = basic_info.loc[cb_id, '转股价']
                Maturity = datetime.strftime(basic_info.loc[cb_id, '到期日期'], '%Y/%m/%d')
                ConvertStart = 5.5
                coupon_info = basic_info.loc[cb_id, '票面利率说明']
                coupon = [float(x) for x in re.findall(r"\d+\.?\d*", coupon_info)]
                coupon[-1] = basic_info.loc[cb_id, '到期赎回价']

                recall = [ConvertStart,
                          basic_info.loc[cb_id, '赎回触发计算时间区间'].astype(int),
                          basic_info.loc[cb_id, '赎回触发计算最大时间区间'].astype(int),
                          basic_info.loc[cb_id, '赎回触发比例'].astype(int)]

                resell = [(basic_info.loc[cb_id, '相对回售期'] / 24).astype(int),
                          basic_info.loc[cb_id, '回售触发计算时间区间'].astype(int),
                          basic_info.loc[cb_id, '回售触发计算时间区间'].astype(int),
                          basic_info.loc[cb_id, '回售触发比例'].astype(int),
                          103]

                term_CB = {
                    "ConvPrice": ConvPrice,
                    "Maturity": Maturity,
                    "ConvertStart": ConvertStart,
                    "Coupon": coupon,
                    "Recall": recall,
                    "Resell": resell
                }
                term_CB['Recall'][-1] *= term_CB['ConvPrice'] / 100
                term_CB['Resell'][-2] *= term_CB['ConvPrice'] / 100
                term_CB['Resell'][-1] *= term_CB['ConvPrice'] / 100

                now_date = '2021/6/1'
                ticker = basic_info.loc[cb_id, '正股代码'].split('.')[0]

                cb_value = CbValuation(date=now_date, stock_ticker=ticker)
                vol, stock_price = cb_value.vol, cb_value.stock_price

                credit = basic_info.loc[cb_id, '发行信用等级']
                year_to_maturity = \
                    (datetime.strptime(Maturity, '%Y/%m/%d') - datetime.strptime(now_date, '%Y/%m/%d')).days // 365
                year_to_maturity = max(year_to_maturity, 1)

                rate = self.bond_ytm[self.bond_ytm['信用评级'] == credit][year_to_maturity].values[0] / 100

                prepared_data[cb_id] = {
                    "stock_price": stock_price,
                    "term_CB": term_CB,
                    "now_date": now_date,
                    "vol": vol,
                    "rate": rate
                }

            self.prepared_data = prepared_data

        with open("C:\\Users\\kai.zhang\\Desktop\\可转债\\cb_params.pkl", "wb") as f:
            pickle.dump(self.prepared_data, f)

    def _load_cb_price(self):
        basic_info = self.basic_info.copy()
        basic_info = basic_info[basic_info['发行信用等级'] != 'A']
        cb_id_list = basic_info['证券代码'].tolist()

        res = w.wsd(','.join(cb_id_list), "close", "2021-06-01", "2021-06-01", "")

        if res.ErrorCode != 0:
            data = pd.DataFrame()
            print("fetch cb data error")
        else:
            if len(res.Data) == 1:
                data = pd.DataFrame(res.Data[0], index=res.Codes, columns=res.Times)
            else:
                data = pd.DataFrame(res.Data, index=res.Codes, columns=res.Times)
            data.columns = ['close']

        self.cb_price = data

    def calculate(self):
        bs_list = []
        tree_list = []
        mc_list = []
        for key, value in tqdm(self.prepared_data.items()):
            stock_price, term_CB, now_date, vol, rate = \
                value['stock_price'], value['term_CB'], value['now_date'], value['vol'], value['rate']

            # B-S估值
            straightBondValue, call_value = cbPricingBlackScholes(stock_price, term_CB, now_date, vol, rate)
            bs_list.append(straightBondValue + call_value)
            # 二叉树估值
            tree_list.append(cbPricingTree(stock_price, term_CB, now_date, vol, rate))
            # MC估值
            mc_list.append(cbPricingMC(stock_price, term_CB, now_date, vol, rate))

        valuation_df = pd.DataFrame({"cb_id": list(self.prepared_data.keys()), "B-S": bs_list,
                                     "tree": tree_list, "mc": mc_list})
        valuation_df = pd.merge(valuation_df, self.cb_price, left_on='cb_id', right_index=True)
        valuation_df.to_csv('C:\\Users\\kai.zhang\\Desktop\\可转债\\可转债估值结果.csv', index=False)


if __name__ == '__main__':
    # # 可转债的基本资料
    # term_CB = {
    #     "ConvPrice": 10.66,
    #     "Maturity": "2021/2/1",
    #     "ConvertStart": 5.5,
    #     "Coupon": [0.2, 0.5, 1.0, 1.5, 1.5, 106.6],
    #     "Recall": [5.5, 15, 30, 130],
    #     "Resell": [2, 30, 30, 70, 103]
    # }
    #
    # now_date = '2015/9/22'
    # ticker = '601727'
    #
    # cb_value = CbValuation(date=now_date, stock_ticker=ticker)
    # vol, stock_price = cb_value.vol, cb_value.stock_price
    # # 同等级/期限的企业债收益率
    # rate = 0.029659
    #
    # cb_price_bs = cbPricingBlackScholes(stock_price, term_CB, now_date, vol, rate)
    # cb_price_tree = cbPricingTree(stock_price, term_CB, now_date, vol, rate)
    # term_CB['Recall'][-1] *= term_CB['ConvPrice'] / 100
    # term_CB['Resell'][-2] *= term_CB['ConvPrice'] / 100
    # term_CB['Resell'][-1] *= term_CB['ConvPrice'] / 100
    # cb_price_mc = cbPricingMC(stock_price, term_CB, now_date, vol, rate)

    cb_info = pd.read_excel("C:\\Users\\kai.zhang\\Desktop\\可转债\\可转债基本资料.xlsx").dropna()
    ytm_data = pd.read_excel("C:\\Users\\kai.zhang\\Desktop\\可转债\\企业债到期收益率.xlsx")
    CbValuationTest(basic_info=cb_info, bond_ytm=ytm_data).calculate()


# -*- coding: utf-8 -*-

import cvxpy as cp
import numpy as np
import pandas as pd


class BL:
    def __init__(self, asset_list, ret, cov, risk_aversion, lb_list, ub_list, total_weight):
        self.asset_list = asset_list
        self.ret = np.matrix(ret)
        self.cov = np.matrix(cov)
        self.risk_aversion = risk_aversion
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.total_weight = total_weight

    def solve(self):
        n = len(self.asset_list)
        w = cp.Variable(n)
        ret = self.ret.T * w
        risk = cp.quad_form(w, self.cov)
        prob = cp.Problem(cp.Maximize(ret - 0.5 * self.risk_aversion * risk),
                          [cp.sum(w) == self.total_weight,
                           w >= self.lb_list,
                           w <= self.ub_list])
        prob.is_dcp()
        prob.solve()
        if prob.status != 'optimal':
            weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * (len(self.asset_list) + 1))
            status = 'infeasible'
        else:
            arr_x = w.value
            for i in range(len(self.asset_list)):
                if arr_x[i] < self.lb_list[i]:
                    arr_x[i] = self.lb_list[i]
            for i in range(len(self.asset_list)):
                if arr_x[i] > self.ub_list[i]:
                    arr_x[i] = self.ub_list[i]
            if sum(arr_x) != self.total_weight:
                arr_x = arr_x / arr_x.sum() * self.total_weight

            weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
            status = 'optimal'
        return weight, status


import datetime
import shlex

import numpy as np
import pandas as pd
from hbshare.fe.XZ import functionality
from hbshare.fe.XZ import db_engine as dben


class Projects:

    def __init__(self):

        from hbshare.fe.XZ.config import config_pfa as config
        self.config_pfa=config.Config()

    # 1.资产配置时序：A股 / 港股 / 债券 / 基金 / 现金等类别时序的累计区域图；
    def asset_allocation_pic(self,plot,asset_allocation_df):

        # 1.资产配置时序：A股 / 港股 / 债券 / 基金 / 现金等类别时序的累计区域图；
        plot.plotly_area(asset_allocation_df[['活期存款', '债券', '基金', '权证', '其他', 'A股',
                                              '港股', '日期']], '资产配置时序')

    # 2.行业配置时序：基于申万一级行业分别规则，对持仓的行业分布做一个时序统计；
    def industry_allocation_pic(self,plot,ind_hld):
        plot.plotly_area(ind_hld, '行业配置时序')

    # 3. 行业集中度时序：持仓前一 / 三 / 五大行业的权重时序
    def industry_centralization_pci(self,plot,indutry_allocation_df_ranked):

        # 3. 行业集中度时序：持仓前一 / 三 / 五大行业的权重时序
        inputdf = indutry_allocation_df_ranked.copy()
        cols = inputdf.columns.tolist()
        cols.remove('日期')
        for col in cols:
            inputdf[col] = [x[0].sum() for x in inputdf[col]]
        plot.plotly_line(inputdf, '行业集中度时序')

    # 4.重仓明细：时序上各个时间点的持仓明细（表格）；
    def stock_holding_pic(self,plot,stk_hld):
        # 4.重仓明细：时序上各个时间点的持仓明细（表格）；
        tempdf = stk_hld.drop('日期',axis=1).T
        input_df=pd.DataFrame()
        for date in tempdf.columns:
            input_df[date]=tempdf[date].sort_values(ascending=False).index[0:20]\
                           +'( '+tempdf[date].sort_values(ascending=False).values[0:20].astype(str)+' )'

        plot.plotly_table(input_df, 5000, '持仓明细')

    # 5.持股集中度时序：持仓前三 / 五 / 十大权重和的时序折线图；
    def stock_centralization_pic(self,plot,asset_allocation_df_ranked):
        # 5.持股集中度时序：持仓前三 / 五 / 十大权重和的时序折线图；
        inputdf = asset_allocation_df_ranked.copy()
        cols = inputdf.columns.tolist()
        cols.remove('日期')
        for col in cols:
            inputdf[col] = [x[0].sum() for x in inputdf[col]]
        plot.plotly_line(inputdf, '持股集中度时序')

    # 6.组合中A股持仓的平均PE / PB / 股息率的折线图；
    def valuation_pic(self,plot,fin_hld,stk_hld,date_list):
        # 6.组合中A股持仓的平均PE / PB / 股息率的折线图；
        tempdf = pd.DataFrame()
        for item in ['PE', 'PB', 'DIVIDENDRATIO']:
            tempdf[item] = ((fin_hld.loc[:, (item, slice(None))].values) * (stk_hld.drop('日期', axis=1).values)).sum(
                axis=1) / stk_hld.drop('日期', axis=1).values.sum(axis=1)
        tempdf['日期'] = date_list
        plot.plotly_line_multi_yaxis(tempdf, '持股估值分析', ['PB', 'DIVIDENDRATIO'])

    # 7.持股分布：所持股票在各类宽基成分股中的权重时序的折线图，包括：沪深300、中证500、中证1000、1800以外。
    def hodling_in_benchmark(self,plot,ben_hld):
        # 7.持股分布：所持股票在各类宽基成分股中的权重时序的折线图，包括：沪深300、中证500、中证1000、1800以外。
        plot.plotly_area(ben_hld, '宽基成分股配置走势')

    # 8 单只基金Alpha R_Square 散点图
    def alpha_rsquare_pic(self,plot,fun,df1,date_list,Jydb):

        #read the factors information from hb database
        sec_code_A=df1['Stock_code'].dropna()[~df1['Stock_code'].dropna().str.contains('H')].unique()
        raw_factors=Jydb.extract_factors(sec_code_A,date_list)

        #calcualte the risk exposure for the fund as a whole
        fund_factors=fun.fund_risk_exposure(fun.purified_stk_hld[['Stock_code','Name','Weight','Stamp_date']].dropna(),
                                            raw_factors,
                                            ['Stamp_date','Stock_code','Weight'])

        #get the return of benchmark index as factors
        index_list=['000905','000300']
        factors_benchmark=Jydb.extract_benchmark_return(date_list,index_list+['000012'])
        factors_benchmark=fun.aggregate_by(factors_benchmark,['TJYF','ZQDM'],'sum','HB')

        #join the classical risk factors with benchmark factors
        fund_factors=pd.merge(fund_factors,factors_benchmark,how='inner',left_on='JYYF',right_on='日期').drop('日期',axis=1)
        factors_list=fund_factors.columns.drop(['JYYF']+index_list).tolist()

        #get the return of the fund by valuation table as the y input for the ols model
        ret_df=fun.generate_ret_df()
        ret_df['JYYF'] = [x.split('-')[0] + x.split('-')[1] for x in ret_df['Stamp_date'].astype(str)]
        ret_df.drop('Stamp_date',axis=1,inplace=True)
        ret_df['Return']=ret_df['Return'].shift(-1)
        ret_df.drop(ret_df.index[-1], axis=0, inplace=True)

        #combine all the data for ols in one dataframe
        ols_df=pd.merge(fund_factors,ret_df,how='inner',left_on='JYYF',right_on='JYYF')
        ols_df['const']=1
        del fund_factors,ret_df,raw_factors

        #8 单只基金Alpha R_Square 散点图

        for num_factors in [6]:
            # pick 6 factors for each round of run as suggested by the research paper of haitong security
            summary_df,simulated_df,selected_factors=fun.calculate_alpha(ols_df, factors_list, num_factors,
                                                                         ret_col= 'Return',date_col='JYYF',
                                                                         bench_factor=index_list)

            plot.plotly_scatter(summary_df[['alpha','rsquar']],'因子组合Alpha_R_square散点图',fix_range=True)

            plot.plotly_line(simulated_df[['simulated_ret', 'real_ret', '日期']], '基金多因子模型拟合效果_因子：{0}'.format(selected_factors))
            # plot.plotly_scatter(simulated_df[['simulated_ret','real_ret']].sort_values('real_ret'), '基金多因子模型拟合效果')

    def prv_hld_analysis (self,prd_name,tablename,pic_list=None):

        ########################################Collecting data########################################
        config=self.config_pfa

        # initial the temp database class
        Pfdb = dben.PrvFunDB()

        # write fold files to DB
        # Pfdb.write2DB(table_name='prvfund',fold_dir="E:\私募主观\主观股多估值表基地",increment=0)

        # initial the class
        Jydb = dben.HBDB()

        # extra data from the temp database
        df1 = Pfdb.extract_from_db(prd_name=prd_name, columns='*', tablename=tablename)

        # get the unique date list and stock code list
        date_list = df1['Stamp_date'].dropna().unique()
        sec_code = df1['Stock_code'].dropna().unique()

        # extra industry data from the JY database
        df2 = Jydb.extract_industry(sec_code)

        # extra financial data from the JY database
        df3 = Jydb.extract_fin_info(sec_code,date_list)

        # extra benchmark data from the JY database
        df4 = Jydb.extract_benchmark(config.Index_type, sec_code=sec_code,date_list=date_list)

        fun = functionality.Untils(df1[['Code','Name','Weight','Stamp_date','Stock_code']], df2, df3, df4)

        asset_allocation_df = fun.asset_allocation_stats()

        ind_hld = fun.aggregate_by(fun.purified_stk_hld, groupby=['Stamp_date', 'FIRSTINDUSTRYNAME'], method='sum',
                                   method_on='Weight')
        indutry_allocation_df_ranked = fun.rank_filter(ind_hld, [1, 3, 5])

        stk_hld = fun.aggregate_by(fun.purified_stk_hld, groupby=['Stamp_date', 'Name'], method='sum',
                                   method_on='Weight')
        asset_allocation_df_ranked = fun.rank_filter(stk_hld, [3, 5, 10])

        fin_hld = fun.aggregate_by(fun.purified_stk_hld, groupby=['Stamp_date', 'Name'], method='sum',
                                   method_on=['PE', 'PB', 'DIVIDENDRATIO'])
        ben_hld = fun.aggregate_by(fun.bench_info, groupby=['Stamp_date', 'Index_type'], method='sum',
                                   method_on='Weight')

        ########################################Drawing Pictures ########################################

        # drawing the pics ,initial the instant
        plot = functionality.Plot(fig_width=1200, fig_height=600)

        if(pic_list is None):

            self.asset_allocation_pic(plot,asset_allocation_df)
            self.industry_allocation_pic(plot,ind_hld)
            self.industry_centralization_pci(plot, indutry_allocation_df_ranked)
            self.stock_holding_pic(plot, stk_hld)
            self.stock_centralization_pic(plot, asset_allocation_df_ranked)
            self.valuation_pic(plot, fin_hld, stk_hld, date_list)
            self.hodling_in_benchmark(plot,ben_hld)
            self.alpha_rsquare_pic(plot, fun, df1, date_list, Jydb)

        else:
            pic_list=[x+'_pic' for x in pic_list]

            if('asset_allocation_pic' in pic_list):
                self.asset_allocation_pic(plot, asset_allocation_df)
            if ('industry_allocation_pic' in pic_list):
                self.industry_allocation_pic(plot,ind_hld)
            if ('industry_centralization_pci' in pic_list):
                self.industry_centralization_pci(plot, indutry_allocation_df_ranked)
            if ('stock_holding_pic' in pic_list):
                self.stock_holding_pic(plot, stk_hld)
            if ('stock_centralization_pic' in pic_list):
                self.stock_centralization_pic(plot, asset_allocation_df_ranked)
            if ('valuation_pic' in pic_list):
                self.valuation_pic(plot, fin_hld, stk_hld, date_list)
            if ('hodling_in_benchmark' in pic_list):
                self.hodling_in_benchmark(plot,ben_hld)
            if ('alpha_rsquare_pic' in pic_list):
                self.alpha_rsquare_pic(plot, fun, df1, date_list, Jydb)

    def prv_fof_analysis(self,fund_code):

        hbdb=dben.HBDB()

        sql = """ 
        select d.*,b.jzrq,b.jjjz,b.ljjz  
        from 
        (
        select a.jjdm,a.jjjc,a.cpfl,a.jjfl,c.mjjdm,c.lhbz,c.sfzz,c.fofjjdl,c.fofejfl,c.tzmb,c.tzfw,c.tzcl,c.tzbz,c.fxsytz 
        from st_hedge.t_st_jjxx a, st_hedge.t_st_sm_jjxx c 
        where a.jjdm=c.jjdm and a.jjdm='{0}'
        ) d left join st_hedge.t_st_jjjz b on d.jjdm=b.jjdm """.format(fund_code)

        hld_prv=hbdb.db2df(sql,db='highuser')

        sql="select jzdm,yxj,moddt_etl from st_hedge.t_st_sm_jjbjjz a where jjdm='{}' ".format(fund_code)
        benchmark_code=hbdb.db2df(sql,db='highuser')

        print('Done')

    def fund_alpha_analysis(self,fund_list):

        fun=functionality.Untils()
        ########################################Collecting data########################################
        hbdb=dben.HBDB()
        for fund in fund_list:

            #read the holding info for the fund
            sql="select jjdm,jsrq,zqdm,zqmc,zjbl from st_fund.t_st_gm_gpzh where jjdm='{0}'".format(fund)
            fund_hld=hbdb.db2df(sql,db='funduser')

            #read the net value info for the fund
            sql="select jjdm,tjyf,hb1y from st_fund.t_st_gm_yhb where jjdm='{0}'".format(fund)
            ret_df=hbdb.db2df(sql,db='funduser')
            ret_df['hb1y']=ret_df['hb1y'] / 100
            ret_df['tjyf']=ret_df['tjyf'].astype(str)


            # read the factors information from hb database
            sec_code= fund_hld['zqdm'][~fund_hld['zqdm'].dropna().str.contains('H')].unique()
            date_list=fund_hld['jsrq'].unique()
            raw_factors = hbdb.extract_factors(sec_code, date_list)

            industry=False
            if(industry==True):
                #add the industry factors if necessary, i am not using the industry factors in the st_ashare.r_st_barra_style_factor
                #because it contains too many industries i am using the 中证指数行业分类2016版 which is less to reduce the computing burder

                sql = """
                select a.SecuCode,b.FirstIndustryName,b.XGRQ from HSJY_GG.SecuMain a 
                left join HSJY_GG.LC_ExgIndustry b on a.CompanyCode=b.CompanyCode  
                where Standard=28 and a.SecuCode in({0}) """.format("'"+"','".join(sec_code)+"'")
                industry_factors=hbdb.db2df(sql,db='readonly').sort_values('XGRQ').drop(['ROW_ID','XGRQ'],axis=1)
                industry_factors.drop_duplicates('SECUCODE',keep='last',inplace=True)

                # sql = """
                # select a.SecuCode,b.FirstIndustryName,b.UpdateTime from HSJY_GG.SecuMain a
                # left join HSJY_GP.LC_STIBExgIndustry b on a.CompanyCode=b.CompanyCode
                # where Standard=28 and a.SecuCode in({0}) """.format("'"+"','".join(sec_code)+"'")
                # industry_factors2=hbdb.db2df(sql,db='readonly').sort_values('UpdateTime').drop(['ROW_ID','UpdateTime'],axis=1)
                # industry_factors2.drop_duplicates('SECUCODE',keep='last',inplace=True)

                industry_factors=pd.concat([industry_factors,pd.get_dummies(industry_factors['FIRSTINDUSTRYNAME'])],axis=1).drop('FIRSTINDUSTRYNAME',axis=1)

                raw_factors=pd.merge(raw_factors,industry_factors,how='left',left_on='ticker',right_on=['SECUCODE']).drop('SECUCODE',axis=1)

            # calcualte the risk exposure for the fund as a whole
            fund_factors = fun.fund_risk_exposure(fund_hld,raw_factors,['jsrq','zqdm','zjbl'])

            jjfl=hbdb.db2df("select jjfl from st_fund.t_st_gm_flls where jjdm='{0}' ".format(fund),db='funduser')['jjfl'][0]
            jzdm_list=hbdb.db2df("select jzdm from st_hedge.t_st_sm_jjflbjjz where jjfl='{}' and jzdm not like 'H%%' ".format(jjfl),db='highuser')['jzdm']\
                .values.tolist()

            sql="select hb1y,zqdm,tjyf from st_market.t_st_zs_yhb  where zqdm in ({0}) ".format(','.join(jzdm_list+['000012']))
            benchmark_df=hbdb.db2df(sql,db='alluser')
            benchmark_df=benchmark_df[benchmark_df['hb1y']!=999]
            benchmark_df=benchmark_df.groupby(['zqdm', 'tjyf']).sum('hb1y').unstack().T
            benchmark_df['日期']=benchmark_df.index.get_level_values(1).astype(str)

            fund_factors=pd.merge(fund_factors,benchmark_df,how='inner',left_on='JYYF',right_on='日期').drop('日期',axis=1)

            factors_list=fund_factors.columns.drop(['JYYF']+jzdm_list).tolist()

            #standardlize the factors to be mean 1 std 1
            for col in factors_list:
                mean=fund_factors[col].mean()
                std=fund_factors[col].std()
                fund_factors[col]=fund_factors[col]/std+1-mean

            # combine all the data for ols in one dataframe
            ols_df = pd.merge(fund_factors, ret_df, how='inner', left_on='JYYF', right_on='tjyf')
            ols_df['const'] = 1
            del fund_factors, ret_df, raw_factors

            plot = functionality.Plot(fig_width=1200, fig_height=600)

            for num_factors in [6]:
                # pick 6 factors for each round of run as suggested by the research paper of haitong security
                summary_df, simulated_df,selected_factors = fun.calculate_alpha(ols_df, factors_list, num_factors, ret_col='hb1y',
                                                               date_col='JYYF',bench_factor=jzdm_list)

                fund_name=hbdb.db2df("select jjjc from st_fund.t_st_gm_jjxx where jjdm='{0}'".format(fund),db='funduser').values[0]
                plot.plotly_scatter(summary_df[['alpha', 'rsquar']], '因子组合Alpha_R_square散点图_{0}'.format(fund_name), fix_range=True)

                plot.plotly_line(simulated_df[['simulated_ret', 'real_ret', '日期']], '基金多因子模型拟合效果_{0}_因子：{1}'.format(fund_name,selected_factors))

    def read_riskaversionfromdb(self,date):

        sql = "select risk_aversion from bl_risk_aversion where date='{0}'".format(date)
        risk_aversion=pd.read_sql(sql,con=dben.PrvFunDB().engine)
        return risk_aversion['risk_aversion'].values[0]

    def bl_model_data_preparation(self,end_date,version):


        from ..AAM.blmodel import  blmodel

        blm=blmodel.BL_Model()
        localdb_engine=dben.PrvFunDB().engine

        #get the assets from pool
        sql=" select * from bl_assets_pool where version='{0}' ".format(version)
        assets_df=pd.read_sql(sql, con=localdb_engine)
        index_assets=assets_df[assets_df['asset_type']=='index']['code'].values.tolist()
        public_funds = assets_df[assets_df['asset_type'] == 'public_fund']['code'].values.tolist()

        blm.bl_model_index_data_preparation(end_date=end_date,asset_list=index_assets,version=version)
        blm.bl_model_publicfund_data_preparation(end_date=end_date, asset_list=public_funds,version=version)

    def bl_model(self,end_date,asset_list,version,tau,ub,lb,asset_type,risk_aversion):

        from ..AAM.blmodel import blmodel

        blm = blmodel.BL_Model()
        engine = dben.PrvFunDB().engine

        # get the cov matrix from local database
        cov_matrix = blm.read_cov_fromdb(engine, asset_list, version, end_date,asset_type)

        # the cov matrix of average return is a shrink of the cov matrix of return itself
        cov_matrix = tau * cov_matrix.values

        # get prio_return from local db
        prio_return = blm.read_return_fromdb(asset_list, engine, version, end_date, 'bl_implied_return', 'implied_ret')

        # get the view_ret from local db
        view_ret = blm.read_return_fromdb(asset_list, engine, version, end_date, 'bl_view_return', 'view_ret')

        views = np.eye(len(asset_list))

        # calculate the view confidence if not read from database directlly
        view_confidence = np.dot(np.dot(views, cov_matrix), views.T) * np.eye(len(views))

        # set upper and lower bound

        constrains = blm.set_boundary(len(asset_list), lb, ub)

        opt_weight = blm.blm_solver(sigma=cov_matrix, mu=prio_return, P=views, Q=view_ret,
                                    Omega=view_confidence, delta=risk_aversion, constrains=constrains)

        opt_restul = pd.DataFrame()
        opt_restul['Asset'] = asset_list
        opt_restul['Weight'] = opt_weight

        return opt_restul

    def fund_classification(self):

        from ..Fund_classifier import classification
        fc = classification.Classifier_Ml()
        #fc.wind_risk_data2localdb(asofdate='2016-01-01')

        #
        # fc.wind_theme_data2localdb(asofdate='2016-01-01')
        #
        #fc.wind_stock_style_data2localdb(asofdate='2016-01-01')
        #
        # fc.model_generation_style()
        #
        # fc.model_generation_theme()
        #
        fc.model_generation_style(value_style='hld')

        fc.model_generation_theme(value_style='hld')
        #
        # fc.model_generation_risk_level()

        # #['2021-12-31','2021-09-30','2021-06-30','2021-03-31','2020-12-31',
        #                  '2020-09-30','2020-06-30','2020-03-31']
        #

        for asofdate in ['2021-12-31','2021-09-30','2021-06-30','2021-03-31','2020-12-31',
                         '2020-09-30','2020-06-30','2020-03-31']:
            fc=classification.Classifier_Ml(asofdate=asofdate)
            #fc.classify()
            fc.classify_hldbase()
            print("{0} done ".format(asofdate))

    def style_distribution(self,plot,table,style_list,clbz='全部'):

        map_dict=dict(zip(['R1','R2','R3','R4','R5','D1','D2','D3','D4','D5'],['低风险','较低风险','中风险','较高风险','高风险','低回撤','较低回撤','中回撤','较高回撤','高回撤',]))

        if(clbz=='全部'):
            sql="select * from {0}".format(table)
        else:
            sql = "select * from {1} where clbz='{0}'".format(clbz,table)

        test = pd.read_sql(sql, con=dben.PrvFunDB().engine)
        date_list=test['style_updated_date'].unique().tolist()

        for col in style_list:

            sql="select distinct {0} from {1} ".format(col,table)
            unique_items= pd.read_sql(sql, con=dben.PrvFunDB().engine)[col].tolist()
            unique_items.remove('')
            plot_df = pd.DataFrame()
            plot_df['items'] = unique_items
            for date in date_list:
                tempdf=test[test['style_updated_date']==date].groupby(col).count()
                tempdf['items']=tempdf.index
                plot_df=pd.merge(plot_df,tempdf[['items','jjdm']],how='left',on='items').fillna(0)
                plot_df.rename(columns={'jjdm':date},inplace=True)
                plot_df[date]=plot_df[date]/plot_df[date].sum(axis=0)*100
            plot_df.sort_values('items',inplace=True)
            if(col=='risk_level' or col=='draw_back_level'):
                plot_df['items']=[map_dict[x] for x in plot_df['items']]
            plot_df.set_index('items',drop=True,inplace=True)
            plot.plotly_style_bar(plot_df.T,'{}分布'.format(col))

    def fund_classification_advance(self):
        from ..Fund_classifier import classification
        #fc=classification.Classifier()
        #fc.classify_threshold(20)
        fc2=classification.Classifier_brinson()
        fc2.classify_socring()

    def brinson_score_pic(self,jjdm):

        sql="select * from brinson_score where jjdm='{}'".format(jjdm)
        scoredf=pd.read_sql(sql,con=dben.PrvFunDB().engine)
        plot=functionality.Plot(fig_width=1000,fig_height=600)

        new_name=['jjdm','交易能力_短期','交易能力_长期','行业配置能力_短期',
                  '行业配置能力_长期','选股能力_短期','选股能力_长期','大类资产配置能力_短期',
                  '大类资产配置能力_长期','asofdate']
        scoredf.columns=new_name
        col=['交易能力_短期','交易能力_长期','行业配置能力_短期',
                  '行业配置能力_长期','选股能力_短期','选股能力_长期','大类资产配置能力_短期',
                  '大类资产配置能力_长期']

        plot.ploty_polar(scoredf[col],'Brinson能力图')

    def fund_classification_barra(self):

        from ..Fund_classifier import classification
        fc=classification.Classifier_barra()
        fc.data_preparation()
        # for jjdm in ['000362','005821','009636','001487']:
        #     fc.classify(jjdm=jjdm,start_date='20181231',end_date='20211231')

    def mutual_fund_concept(self,asofdate):

        engine=dben.PrvFunDB().engine

        q3df = pd.read_excel(r"E:\GitFolder\hbshare\fe\Fund_classifier\重仓持股(汇总).xlsx")[
            ['代码', '名称', '持股总量(万股)', '季报持仓变动(万股)', '持股市值占基金净值比(%)']]
        target_list = q3df['代码'].fillna('0').tolist()
        q3df.fillna(0)
        q3df = q3df[q3df['持股总量(万股)'] > 0]

        sql = "select * from stock_concept where 证券代码 in ({0})".format("'" + "','".join(target_list) + "'")
        conceptdf = pd.read_sql(sql, con=engine).fillna('')

        df = pd.merge(q3df, conceptdf, how='inner', left_on='代码', right_on='证券代码')
        concept_listdf = pd.read_excel(r"E:\GitFolder\hbshare\fe\Fund_classifier\概念分类.xlsx")
        concept_class = concept_listdf.columns
        csv_df = pd.DataFrame()
        for cls in concept_class:
            con_weight = []
            con_change = []
            con_total = []
            outputdf = pd.DataFrame()
            concept_list = concept_listdf[concept_listdf[cls].notnull()][cls].astype(str)
            for concept in concept_list:
                try:
                    df[concept] = df['所属概念板块[交易日期] 最新收盘日'].str.contains(concept).astype(int)
                    df[concept + '_w'] = df[concept] * df['持股市值占基金净值比(%)']
                    df[concept + '_n'] = df[concept] * df['季报持仓变动(万股)']
                    df[concept + '_t'] = df[concept] * df['持股总量(万股)']
                    con_weight.append(df[concept + '_w'].sum(axis=0))
                    con_change.append(df[concept + '_n'].sum(axis=0))
                    con_total.append(df[concept + '_t'].sum(axis=0))
                except Exception:
                    print(concept)
                    concept_list.remove(concept)
                    continue

            outputdf[cls] = concept_list
            outputdf['持仓权重(%)_' + cls] = con_weight
            outputdf['季报持仓变动(万股)'] = con_change
            outputdf['持股总量(万股)'] = con_total

            outputdf['上季度持股总量(万股)'] = outputdf['持股总量(万股)'] - outputdf['季报持仓变动(万股)']
            outputdf['持股变动%_' + cls] = outputdf['季报持仓变动(万股)'].fillna(0) / outputdf['上季度持股总量(万股)'] * 100
            outputdf.loc[outputdf['持股变动%_' + cls]==np.inf,'持股变动%_' + cls]='新增'
            outputdf.drop(['上季度持股总量(万股)', '季报持仓变动(万股)', '持股总量(万股)'], axis=1, inplace=True)
            outputdf.sort_values(['持仓权重(%)_' + cls, '持股变动%_' + cls], inplace=True, ascending=False)
            outputdf = outputdf[outputdf['持仓权重(%)_' + cls] > 0].reset_index(drop=True)
            csv_df = pd.concat([csv_df, outputdf], axis=1)

        csv_df['asofdate'] = asofdate
        #check if data already exist
        sql='select distinct asofdate from mutual_fund_highlighthld_concept'
        asofdatelist=pd.read_sql(sql,con=engine)['asofdate']
        if(asofdate in asofdatelist):
            sql="delete from mutual_fund_highlighthld_concept where asofdate='{0}'".format(asofdate)
            engine.execute(sql)
        csv_df.to_sql('mutual_fund_highlighthld_concept', con=engine,index=False,if_exists='append')
        print('The concept distribution data has been inserted into tbale mutual_fund_highlighthld_concept')

    def mutual_fund_summary_20220115(self,date):

        outputdf=pd.DataFrame()

        theme_map={'大金融' : ['银行','券商','房地产','保险','非银金融'],
                   '消费' : ['美容护理','家用电器','酒类','制药','医疗保健','生物科技','商业服务','零售','纺织服装','食品','农业','家居用品','餐饮旅游','软饮料','医药生物','商业贸易','商贸零售','食品饮料','农林牧渔','休闲服务','纺织服饰'],
                   'TMT' : ['半导体','电子元器件','电脑硬件','软件','互联网','文化传媒','电子','计算机','传媒','通信'],
                   '周期': ['采掘','有色金属','化工原料','基本金属','贵金属','钢铁','化纤','建筑','煤炭','化肥农药','石油天然气','日用化工','建材','石油化工','石油石化','化工','基础化工','黑色金属'],
                   '制造' : ['精细化工','建筑材料','工业机械','电工电网','电力','发电设备','汽车零部件','航天军工','能源设备','航空','环保','汽车','通信设备','海运','工程机械','国防军工','电力设备','电气设备','机械设备','建筑装饰','公用事业','环保','交通运输','制造','社会服务','轻工制造'],
                   }

        hld_summary=pd.read_excel(r"E:\基金分类\重仓持股(汇总){}.xlsx".format(date))[
            ['代码', '名称', '持股市值占基金净值比(%)']]

        style_ind=pd.read_excel(r"E:\GitFolder\hbshare\fe\Fund_classifier\股票风格行业.xlsx")
        sql="select * from juchao_style where record_date='20220126'"
        style_juchao=pd.read_sql(sql,con=dben.PrvFunDB().engine)
        style_juchao['所属规模风格类型']=[x+'型' for x in style_juchao['所属规模风格类型']]
        style_ind=pd.merge(style_ind,style_juchao[['证券简称','所属规模风格类型']],how='left',on='证券简称')
        style_ind.loc[style_ind['所属规模风格类型_y'].notnull(),'所属规模风格类型_x']\
            =style_ind.loc[style_ind['所属规模风格类型_y'].notnull()]['所属规模风格类型_y']
        style_ind.drop('所属规模风格类型_y',axis=1,inplace=True)
        style_ind.rename(columns={'所属规模风格类型_x':'所属规模风格类型'},inplace=True)

        theme_list=['新能源','碳科技','芯片','锂电池','光伏','新基建','白酒']
        theme_w=[]
        for theme in theme_list:
            tempdf=pd.read_excel(r"E:\GitFolder\hbshare\fe\Fund_classifier\{}.xls".format(theme))
            tempdf=pd.merge(tempdf,hld_summary,how='left',left_on='证券代码',right_on='代码')
            theme_w.append(tempdf['持股市值占基金净值比(%)'].sum())

        outputdf['主题']=theme_list
        outputdf['持仓权重(%)_主题']=theme_w

        style_ind['宽基']=''
        index_list=['沪深300','中证500','中证1000']
        for ind in index_list:
            style_ind.loc[style_ind[ind]=='是','宽基']=ind

        style_ind['宽基2'] = ''
        index_list = ['上证50']
        for ind in index_list:
            style_ind.loc[style_ind[ind] == '是', '宽基2'] = ind


        style_ind.drop(index_list,axis=1,inplace=True)
        style_ind['大类行业']=''
        for big_ind in theme_map.keys():
            style_ind.loc[[x in theme_map[big_ind] for x in style_ind['申万行业']],'大类行业']=big_ind

        style_ind=pd.merge(hld_summary,style_ind,how='left',left_on='代码',right_on='证券代码')

        for col in ['所属规模风格类型','申万行业','宽基','宽基2','大类行业']:

            tempdf=style_ind.groupby(col).sum()
            tempdf.reset_index(drop=False,inplace=True)
            tempdf.rename(columns={'持股市值占基金净值比(%)':"持仓权重(%)_{}".format(col)},inplace=True)
            tempdf.drop(tempdf[tempdf[col] == ''].index, axis=0, inplace=True)
            outputdf=pd.concat([outputdf,tempdf],axis=1)

        outputdf.to_csv('持仓汇总_{}.csv'.format(date),index=False,encoding='gbk')
        print('Done')

    def mutual_fund_concept_heatmap(self,fig_width,fig_height,asofdate):

        plot = functionality.Plot(fig_width=fig_width, fig_height=fig_height)

        sql="select * from mutual_fund_highlighthld_concept where asofdate='{}'"\
            .format(asofdate)
        raw_df=pd.read_sql(sql,con=dben.PrvFunDB().engine)

        col_list=list(set([x.split('_')[-1] for x in  raw_df.columns]))
        item_list=list(set([x.split('_')[0] for x in  raw_df.columns]))
        item_list=list(set(item_list).difference(set(col_list)))
        item_list.sort()
        col_list.remove('asofdate')
        col_list.sort()

        f = lambda x: '%.4f%%' % x

        for col in col_list:
            new_list=[x+'_'+col for x in item_list]
            tempdf = raw_df[raw_df[col].notnull()][[col] + new_list]
            if(len(tempdf)>=9):
                width=3
            elif(len(tempdf)>=4):
                width=2
            tempdf=tempdf.iloc[0:width*width]
            tempdf.sort_values(new_list[0], ascending=False,inplace=True)
            z=tempdf[new_list[0]].astype(float).values.reshape(width,width)
            z_per=tempdf[new_list[0]].astype(float).apply(f).values.reshape(width,width)
            z_text=tempdf[col].values.reshape(width,width)
            z_text=z_text + '： ' +z_per
            plot.ploty_heatmap(z, z_text, col+"_权重")

            tempdf.sort_values(new_list[1], ascending=False,inplace=True)
            z=tempdf[new_list[1]].astype(float).values.reshape(width,width)
            z_per = tempdf[new_list[1]].astype(float).apply(f).values.reshape(width, width)
            z_text=tempdf[col].values.reshape(width,width)
            z_text=z_text + '： ' + z_per
            plot.ploty_heatmap(z,z_text,col+"_增长率")

    def mutual_industry_allocation_analysis(self):

        df1=pd.read_excel(r"E:\基金分类\历次持仓汇总.xlsx")[['代码', '名称', '持股市值占基金净值比(%)','报告期']]
        df1=df1[df1['报告期']>='2010q1']

        hbdb=dben.HBDB()
        sql="select a.zqdm,b.yjxymc,b.xxfbrq from st_ashare.t_st_ag_zqzb a left join st_ashare.t_st_ag_gshyhfb b on a.gsdm=b.gsdm where  b.xyhfbz={0} and a.zqlb=1 "\
            .format(24)
        ind_map=hbdb.db2df(sql,db='alluser')
        ind_map.reset_index(drop=True,inplace=True)
        ind_map.sort_values(['zqdm','xxfbrq'],inplace=True)
        temp=ind_map['zqdm']
        temp.drop_duplicates(keep='last', inplace=True)
        ind_map = ind_map.loc[temp.index][['zqdm', 'yjxymc']]

        df1['zqdm'] = [x[0:6] for x in df1['代码']]
        df1=pd.merge(df1,ind_map,how='left',on='zqdm')
        df1.drop('zqdm',axis=1,inplace=True)
        df1.rename(columns={'yjxymc': '所属行业'}, inplace=True)

        df2=pd.DataFrame()
        for year in range(2010,2021):
            temp1=pd.read_excel(r"E:\基金分类\{0}Q1重仓持股(汇总).xlsx".format(year))
            temp1.columns = temp1.iloc[0]
            temp1=temp1.iloc[1:-2][['代码', '名称', '持股市值占基金净值比(%)']]
            temp1['zqdm']=[x[0:6] for x in temp1['代码']]
            temp1['报告期']=str(year)+'q1'
            temp1=pd.merge(temp1,ind_map,how='left',on='zqdm')
            temp1.rename(columns={'yjxymc':'所属行业'},inplace=True)
            temp1.drop('zqdm',axis=1,inplace=True)

            temp2 = pd.read_excel(r"E:\基金分类\{0}Q3重仓持股(汇总).xlsx".format(year))
            temp2.columns = temp2.iloc[0]
            temp2=temp2.iloc[1:-2][['代码', '名称', '持股市值占基金净值比(%)']]
            temp2['zqdm'] = [x[0:6] for x in temp2['代码']]
            temp2['报告期']=str(year)+'q3'
            temp2=pd.merge(temp2,ind_map,how='left',on='zqdm')
            temp2.rename(columns={'yjxymc':'所属行业'},inplace=True)
            temp2.drop('zqdm', axis=1, inplace=True)

            df2=pd.concat([df2,temp1],axis=0)
            df2 = pd.concat([df2, temp2], axis=0)


        df=pd.concat([df1,df2],axis=0)
        del df1,df2,temp1,temp2
        df=df.sort_values(['报告期', '所属行业'], ascending=False)
        df['hk']=[x.split('.')[1] for x in df['代码']]
        #remove hk stock
        df=df[df['hk']!='HK']
        df1=df.groupby(['所属行业','报告期']).sum()['持股市值占基金净值比(%)']
        df2 = df.groupby(['报告期']).sum()['持股市值占基金净值比(%)']
        df=pd.merge(df1,df2,how='left',left_index=True,right_index=True)
        df['行业占比%'] = df['持股市值占基金净值比(%)_x'] / df['持股市值占基金净值比(%)_y'] * 100
        df = df.sort_index(level=[0, 1])

        inddf=pd.read_excel(r"E:\基金分类\行业统计.xlsx")
        inddf=inddf.iloc[0:-2]
        inddf['日期']=[str(x)[0:10].replace("-","") for x in inddf['日期'] ]
        inddf['月']=[x[4:6] for x in inddf['日期']]
        inddf['年'] = [x[0:4] for x in inddf['日期']]
        inddf.loc[inddf['月']=='03','月']='q1'
        inddf.loc[inddf['月'] == '06', '月'] = 'q2'
        inddf.loc[inddf['月'] == '09', '月'] = 'q3'
        inddf.loc[inddf['月'] == '12', '月'] = 'q4'
        inddf['日期']=inddf['年']+inddf['月']
        inddf=inddf[inddf['日期'].str.contains('q')]
        # inddf.set_index('日期',drop=True,inplace=True)


        zszcol=['总市值(合计)采掘', '总市值(合计)化工', '总市值(合计)钢铁', '总市值(合计)有色金属',
       '总市值(合计)建筑材料', '总市值(合计)建筑装饰', '总市值(合计)电气设备', '总市值(合计)机械设备',
       '总市值(合计)国防军工', '总市值(合计)汽车', '总市值(合计)家用电器', '总市值(合计)纺织服装', '总市值(合计)轻工制造',
       '总市值(合计)商业贸易', '总市值(合计)农林牧渔', '总市值(合计)食品饮料', '总市值(合计)休闲服务',
       '总市值(合计)医药生物', '总市值(合计)公用事业', '总市值(合计)交通运输', '总市值(合计)房地产', '总市值(合计)电子',
       '总市值(合计)计算机', '总市值(合计)传媒', '总市值(合计)通信', '总市值(合计)银行', '总市值(合计)非银金融',
       '总市值(合计)综合']
        pecol=['市盈率(TTM,算术平均)采掘', '市盈率(TTM,算术平均)化工', '市盈率(TTM,算术平均)钢铁',
       '市盈率(TTM,算术平均)有色金属', '市盈率(TTM,算术平均)建筑材料', '市盈率(TTM,算术平均)建筑装饰',
       '市盈率(TTM,算术平均)电气设备', '市盈率(TTM,算术平均)机械设备', '市盈率(TTM,算术平均)国防军工',
       '市盈率(TTM,算术平均)汽车', '市盈率(TTM,算术平均)家用电器', '市盈率(TTM,算术平均)纺织服装',
       '市盈率(TTM,算术平均)轻工制造', '市盈率(TTM,算术平均)商业贸易', '市盈率(TTM,算术平均)农林牧渔',
       '市盈率(TTM,算术平均)食品饮料', '市盈率(TTM,算术平均)休闲服务', '市盈率(TTM,算术平均)医药生物',
       '市盈率(TTM,算术平均)公用事业', '市盈率(TTM,算术平均)交通运输', '市盈率(TTM,算术平均)房地产',
       '市盈率(TTM,算术平均)电子', '市盈率(TTM,算术平均)计算机', '市盈率(TTM,算术平均)传媒',
       '市盈率(TTM,算术平均)通信', '市盈率(TTM,算术平均)银行', '市盈率(TTM,算术平均)非银金融',
       '市盈率(TTM,算术平均)综合']
        pbcol=['市净率(算术平均)采掘', '市净率(算术平均)化工', '市净率(算术平均)钢铁',
       '市净率(算术平均)有色金属', '市净率(算术平均)建筑材料', '市净率(算术平均)建筑装饰', '市净率(算术平均)电气设备',
       '市净率(算术平均)机械设备', '市净率(算术平均)国防军工', '市净率(算术平均)汽车', '市净率(算术平均)家用电器',
       '市净率(算术平均)纺织服装', '市净率(算术平均)轻工制造', '市净率(算术平均)商业贸易', '市净率(算术平均)农林牧渔',
       '市净率(算术平均)食品饮料', '市净率(算术平均)休闲服务', '市净率(算术平均)医药生物', '市净率(算术平均)公用事业',
       '市净率(算术平均)交通运输', '市净率(算术平均)房地产', '市净率(算术平均)电子', '市净率(算术平均)计算机',
       '市净率(算术平均)传媒', '市净率(算术平均)通信', '市净率(算术平均)银行', '市净率(算术平均)非银金融',
       '市净率(算术平均)综合']

        item_name_list=['mkt_w%','pe','pb']
        col_lists=[zszcol,pecol,pbcol]

        for i in range(3):
            col_list=col_lists[i]
            item_name=item_name_list[i]
            temp=inddf[['日期']+col_list]
            values=[]
            name=[]
            date=[]
            if(item_name=='mkt_w%'):
                for i in range(len(temp)):
                    values+=(temp.iloc[i][col_list]/temp.iloc[i][col_list].sum()*100).values.tolist()
                    name+=col_list
                    date+=[temp.iloc[i]['日期']]*len(col_list)
            else:
                for i in range(len(temp)):
                    values+=temp.iloc[i][col_list].values.tolist()
                    name+=col_list
                    date+=[temp.iloc[i]['日期']]*len(col_list)

            tempdf=pd.DataFrame()
            tempdf[item_name]=values
            tempdf['所属行业']=[x.split(')')[1] for x in name]
            tempdf['报告期']=date
            tempdf = tempdf.sort_values(['报告期', '所属行业'], ascending=False)
            tempdf=tempdf.groupby(['所属行业','报告期']).sum()
            tempdf=tempdf.sort_index(level=[0,1])
            df=pd.merge(df,tempdf,how='left',left_index=True,right_index=True)


        df.rename(columns={'持股市值占基金净值比(%)_x':'持股市值占基金净值比(%)',
                           '持股市值占基金净值比(%)_y':'全行业持股市值占基金净值比(%)',
                           'mkt_w%':'全市场占比%','行业占比%':'公募重仓占比%'},inplace=True)

        df['配置系数'] = df['公募重仓占比%'] / df['全市场占比%']
        df.loc[df['配置系数']==np.inf,'配置系数']=np.nan
        df.loc[df['pe'] == 0, 'pe'] = np.nan
        df.loc[df['pb'] == 0, 'pb'] = np.nan
        df.loc[df['全市场占比%'] == 0, '全市场占比%'] = np.nan

        quant_df=pd.DataFrame()
        for indus in df.index.get_level_values(0).unique():
            tempdf=df.loc[indus][['配置系数','公募重仓占比%','pe','pb']].rank()
            tempdf=tempdf/len(tempdf)
            tempdf['所属行业']=indus
            tempdf.reset_index(inplace=True)
            quant_df=pd.concat([quant_df,tempdf],axis=0)

        quant_df=quant_df.groupby(['所属行业', '报告期']).sum()
        for col in quant_df.columns:
            quant_df.rename(columns={col:col+"_分位数"},inplace=True)
        df=pd.merge(df,quant_df,how='left',right_index=True,left_index=True)
        df.to_csv('mutual_fund_change.csv',encoding='gbk')

        return df


if __name__ == '__main__':

    pj=Projects()
    df=pj.mutual_industry_allocation_analysis()






"""
辅助函数模块
"""
import numpy as np


def cal_annual_return(return_series):
    T = len(return_series)
    annual_return = (1 + return_series).prod() ** (52 / T) - 1

    return annual_return


def cal_annual_volatility(return_series):
    vol = return_series.std() * np.sqrt(52)

    return vol


def cal_max_drawdown(nav_series):
    drawdown_series = nav_series / (nav_series.cummax()) - 1

    return drawdown_series.min()


def cal_sharpe_ratio(return_series, rf):
    annual_return = cal_annual_return(return_series)
    vol = cal_annual_volatility(return_series)
    sharpe_ratio = (annual_return - rf) / vol

    return sharpe_ratio


#!/usr/bin/python
#coding:utf-8
import hbshare as hbs
import pandas as pd
import pyecharts.options as opts
from pyecharts.charts import Line
from pyecharts.globals import ThemeType


class PrivateFundRAReturnCalculator:
    def __init__(self, fund_id_list, start_date, end_date):
        self.fund_id_list = fund_id_list
        self.start_date = start_date
        self.end_date = end_date
        self._load_data()

    @staticmethod
    def _load_calendar(start_date, end_date):
        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            start_date, end_date)
        res = hbs.db_data_query('readonly', sql_script)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)

        return df[['calendarDate', 'isOpen', 'isWeekEnd', 'isMonthEnd']]

    def _load_fund_info(self):
        data = []
        for fund_id in self.fund_id_list:
            sql_script = "SELECT JJDM, JJJC, CLRQ, ZZRQ FROM funddb.jjxx1 WHERE JJDM = '{}'".format(fund_id)
            res = hbs.db_data_query('readonly', sql_script)
            df = pd.DataFrame(res['data'])
            data.append(df)
        data = pd.concat(data)

        return data

    def _load_fund_nav(self):
        data = []
        for fund_id in self.fund_id_list:
            sql_script = "SELECT a.jjdm fund_id, b.jzrq tradeDate, b.fqdwjz as adj_nav from " \
                         "fundapp.jjxx1 a, fundapp.smlj b where a.cpfl = '4' and a.jjdm = b.jjdm " \
                         "and a.jjzt not in ('3') " \
                         "and a.jjdm = '{}' " \
                         "order by b.jzrq".format(fund_id)
            res = hbs.db_data_query('readonly', sql_script)
            df = pd.DataFrame(res['data'])
            data.append(df)
        data = pd.concat(data)

        return data

    @staticmethod
    def cal_ret(nav_series):
        period_nav = nav_series.copy().dropna()
        annual_return = (period_nav.iloc[-1] / period_nav.iloc[0]) ** (2/3) - 1
        return annual_return

    def _load_data(self):
        fund_nav = self._load_fund_nav()
        all_ret = []
        for fund_id in self.fund_id_list:
            df = fund_nav[fund_nav['FUND_ID'] == fund_id].set_index('TRADEDATE')['ADJ_NAV'].sort_index()
            range_start, range_end = df.index[0], df.index[-1]
            date_range = [x.strftime("%Y%m%d") for x in pd.date_range(range_start, range_end)]
            df = df.reindex(date_range)
            rolling_ret = df.rolling(548, min_periods=52).apply(lambda x: self.cal_ret(x))[548:]
            all_ret.append(rolling_ret)

        all_ret = pd.concat(all_ret, axis=1).dropna(how='any')
        fund_info = self._load_fund_info()
        all_ret.columns = fund_info['JJJC'].tolist()
        # reindex
        interval_start, interval_end = all_ret.index[0], all_ret.index[-1]
        calendar_df = self._load_calendar(interval_start, interval_end)
        weekend_list = calendar_df[
            (calendar_df['isWeekEnd'] == 1) & (calendar_df['isOpen'] == 1)]['calendarDate'].tolist()
        all_ret = all_ret.reindex(weekend_list)

        all_ret = all_ret[(all_ret.index >= self.start_date) & (all_ret.index <= self.end_date)]

        self.data = all_ret

    def draw_picture(self):
        df = self.data.copy()
        df *= 100

        line = Line(
            init_opts=opts.InitOpts(
                width='1200px',
                height='600px',
                theme=ThemeType.WALDEN
            )
        ).set_global_opts(
            title_opts=opts.TitleOpts(title="滚动18个月年化收益率走势"),
            tooltip_opts=opts.TooltipOpts(trigger="axis", axis_pointer_type="cross"),
            legend_opts=opts.LegendOpts(pos_right="center", pos_top='5%'),
            xaxis_opts=opts.AxisOpts(
                type_='category',
                is_inverse=True,
                axistick_opts=opts.AxisTickOpts(is_show=True, is_align_with_label=True),
                splitline_opts=opts.SplitLineOpts(is_show=True),
                axisline_opts=opts.AxisLineOpts(on_zero_axis_index=False),
            ),
            yaxis_opts=opts.AxisOpts(
                type_="value",
                min_='dataMin',
                axistick_opts=opts.AxisTickOpts(is_show=True),
                splitline_opts=opts.SplitLineOpts(is_show=True),
                axislabel_opts=opts.LabelOpts(formatter="{value} %"),
            ),
            datazoom_opts=[
                opts.DataZoomOpts(range_start=0, range_end=100),
                opts.DataZoomOpts(type_="inside")
            ],
        ).add_xaxis(
            xaxis_data=df.index.tolist()[::-1]
        )

        for fund_name in df.columns:
            line.add_yaxis(
                series_name=fund_name,
                y_axis=df[fund_name].round(2).tolist()[::-1],
                is_symbol_show=False,
                is_smooth=True,
                linestyle_opts=opts.LineStyleOpts(width=1.5),
                label_opts=opts.LabelOpts(is_show=False)
            )

        return line


if __name__ == '__main__':
    results = PrivateFundRAReturnCalculator(
        ['P00107', 'SK8618', 'P00193'], start_date='20190220', end_date='20210301').draw_picture()



"""
可转债定价模型；Monte-Carlo
"""
import numpy as np
# from cb_pricing_BS import run_time
from hbshare.quant.Kevin.convertible_bond.cb_pricing_BS import _pv
from cb_pricing_Tree import _cashFlowDict


def _monteCarlo(stock, vol, r, numNodes, numToMock=10000):
    """模拟风险中性世界下的股价"""
    arrStock = np.ones([numToMock, numNodes + 1])
    dt = 1 / 250.
    rc = np.log(1 + r)

    arrMock = np.cumprod(
        np.exp((rc - 0.5 * (vol ** 2)) * dt + vol * np.sqrt(dt) * np.random.randn(numToMock, numNodes)), axis=1)

    arrStock[:, 1:] = arrMock
    arrStock *= stock

    return arrStock


def _sliceInMC(array, point, length):

    return array[int(max([0, point - length])): int(point)]


def _isRecall(arr, arrTimeSeries, recall_term):
    logicTime = np.array(arrTimeSeries[:]) < recall_term[0]
    logicPrice = arr[:] > recall_term[-1]

    return 1 if np.sum(1 * logicTime * logicPrice) >= recall_term[1] else 0


def _isResell(arr, arrTimeSeries, resell_term):
    logicTime = np.array(arrTimeSeries[:]) < resell_term[0]
    logicPrice = arr[:] < resell_term[-2]

    return 1 if np.sum(1 * logicTime * logicPrice) >= resell_term[1] else 0


def _processRecall(term, row, i, arrTimeSeries):

    if row[i] > term['Recall'][-1]:
        sliceArr = _sliceInMC(row, i, term['Recall'][2])
        sliceArrTimeSeries = _sliceInMC(arrTimeSeries, i, term['Recall'][2])

        isBeRecall = _isRecall(sliceArr, sliceArrTimeSeries, term['Recall'])

        if isBeRecall:
            thisRowEndTime = i / 250.
            thisRowValue = row[i]
        else:
            thisRowEndTime = None
            thisRowValue = None

        return isBeRecall, thisRowEndTime, thisRowValue
    else:
        return None, None, None


def _processResell(term, row, i, arrTimeSeries):

    if row[i] < term['Resell'][-2]:
        sliceArr = _sliceInMC(row, i, term['Resell'][2])
        sliceArrTimeSeries = _sliceInMC(arrTimeSeries, i, term['Resell'][2])

        isBeResell = _isResell(sliceArr, sliceArrTimeSeries, term['Resell'])

        if isBeResell:
            thisRowEndTime = i / 250.
            thisRowValue = term['Resell'][-1]
        else:
            thisRowEndTime = None
            thisRowValue = None

        return isBeResell, thisRowEndTime, thisRowValue
    else:
        return None, None, None


def _pvCashFlowMC(thisRowEndTime, thisRowValue, dictCF, r):

    cfT, cf = [thisRowEndTime], [thisRowValue]

    for time in dictCF:
        if time / 250.0 < thisRowEndTime:
            cfT.append(time / 250.0)
            cf.append(dictCF[time])

    return _pv(cf, cfT, r)


# @run_time
def cbPricingMC(stock, term, now, vol, r, numToMock=10000, isPrint=True):
    """
    MC定价函数
    @param stock: 当前股价
    @param term: 转债条款变量
    @param now: 当前日期
    @param vol: 隐含波动率
    @param r: 无风险利率(同等级期限信用债收益率)
    @param numToMock: 模拟次数
    @param isPrint
    """
    dictCF = _cashFlowDict(term['Coupon'], term['Maturity'], now)

    numNodes = int(max(dictCF))
    arrTimeSeries = [(numNodes - i) / 250.0 for i in range(numNodes)]

    arrMC = _monteCarlo(stock, vol, r, numNodes, numToMock)

    v = []
    j = 0
    k = 0
    for row in arrMC:
        for i in range(numNodes):
            # 是否触发赎回
            isBeRecall, thisRowEndTime, thisRowValue = _processRecall(term, row, i, arrTimeSeries)

            if isBeRecall:
                j += 1
                thisRowValue *= 100 / term['ConvPrice']
                v.append(_pvCashFlowMC(thisRowEndTime, thisRowValue, dictCF, r))
                break
            # 是否触及回售
            isBeResell, thisRowEndTime, thisRowValue = _processResell(term, row, i, arrTimeSeries)

            if isBeResell:
                k += 1
                thisRowValue *= 100 / term['ConvPrice']
                v.append(_pvCashFlowMC(thisRowEndTime, thisRowValue, dictCF, r))
                break

        else:
            # 整个生命周期都没有触发条款，计算到期价值
            thisRowEndTime = numNodes / 250.
            thisRowValue = np.max([row[-1] * 100 / term['ConvPrice'], term['Coupon'][-1]])

            v.append(_pvCashFlowMC(thisRowEndTime, thisRowValue, dictCF, r))

    if isPrint:
        print("MC模型的可转债价值: {}".format(np.mean(v), '0.2f'))
        # print("赎回路径次数: {}".format(j))
        # print("回售路径次数: {}".format(k))

    return np.mean(v)


if __name__ == '__main__':
    # term_CB = {
    #     "ConvPrice": 14.01,
    #     "Maturity": "2025/12/19",
    #     "ConvertStart": 5.5,
    #     "Coupon": [0.5, 0.8, 1.2, 1.8, 2.5, 118],
    #     "Recall": [5.5, 15, 30, 130],
    #     "Resell": [2, 30, 30, 70, 103]
    # }
    #
    # stock_price = 77.4
    # stock_vol = 0.80047
    # rate = 0.067289
    #
    # now_date = '2020/8/21'

    term_CB = {
        "ConvPrice": 10.66,
        "Maturity": "2021/2/1",
        "ConvertStart": 5.5,
        "Coupon": [0.2, 0.5, 1.0, 1.5, 1.5, 106.6],
        "Recall": [5.5, 15, 30, 130],
        "Resell": [2, 30, 30, 70, 103]
    }

    stock_price = 11.99
    # stock_price = stock_price * 100 / term_CB['ConvPrice']
    stock_vol = 0.7847
    rate = 0.029659

    now_date = '2015/9/22'

    term_CB['Recall'][-1] *= term_CB['ConvPrice'] / 100
    term_CB['Resell'][-2] *= term_CB['ConvPrice'] / 100
    term_CB['Resell'][-1] *= term_CB['ConvPrice'] / 100
    cbPricingMC(stock_price, term_CB, now_date, stock_vol, rate)


# -*- coding: utf-8 -*-

from scipy.optimize import minimize
import numpy as np
import pandas as pd


class RiskParity:
    def __init__(self, asset_list, cov, lb_list, ub_list, total_weight):
        self.asset_list = asset_list
        self.omega = np.matrix(cov)
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.total_weight = total_weight

    def _target_func(self, x):
        tmp = np.sqrt((np.matrix(x) * self.omega * np.matrix(x).T).A1)
        risk = np.multiply(np.matrix(x).T, self.omega * np.matrix(x).T).A1 / tmp
        delta_risk = sum([sum((i - risk) ** 2) for i in risk])
        return delta_risk

    def _get_cons(self):
        cons = ({'type': 'eq', 'fun': lambda x: sum(x) - self.total_weight})
        return cons

    def _get_bnds(self):
        x0 = 1 / (np.diag(self.omega) ** 0.5)
        x0 = x0 / sum(x0)
        bnds = tuple((self.lb_list[i], self.ub_list[i]) for i in range(len(x0)))
        return bnds, x0

    def solve(self, ftol=1e-20):
        cons = self._get_cons()
        bnds, x0 = self._get_bnds()
        options = {'disp': False, 'maxiter': 500, 'ftol': ftol}

        res = minimize(self._target_func, x0, bounds=bnds, constraints=cons, method='SLSQP', options=options)
        if not res.success:
            weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * len(self.asset_list + 1))
            status = 'infeasible'

        else:
            arr_x = np.array(list(res.x))
            for i in range(len(self.asset_list)):
                if arr_x[i] < self.lb_list[i]:
                    arr_x[i] = self.lb_list[i]
            for i in range(len(self.asset_list)):
                if arr_x[i] > self.ub_list[i]:
                    arr_x[i] = self.ub_list[i]
            if sum(arr_x) != self.total_weight:
                arr_x = arr_x / arr_x.sum() * self.total_weight

            weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
            status = 'optimal'
        return weight, status


class RiskBudget:
    def __init__(self, asset_list, cov, lb_list, ub_list, total_weight, risk_budget_list):
        self.asset_list = asset_list
        self.omega = np.matrix(cov)
        self.lb_list = lb_list
        self.ub_list = ub_list
        self.total_weight = total_weight
        self.risk_budget_list = risk_budget_list

    def _target_func(self, x):
        tmp = np.sqrt((np.matrix(x) * self.omega * np.matrix(x).T).A1)
        r_risk = np.multiply(np.matrix(x).T, self.omega * np.matrix(x).T).A1 / tmp
        b_risk = self.risk_budget_list * tmp
        delta_risk = sum((b_risk - r_risk) ** 2)
        return delta_risk

    def _get_cons(self):
        cons = ({'type': 'eq', 'fun': lambda x: sum(x) - self.total_weight})
        return cons

    def _get_bnds(self):
        x0 = 1 / (np.diag(self.omega) ** 0.5)
        x0 = x0 / sum(x0)
        bnds = tuple((self.lb_list[i], self.ub_list[i]) for i in range(len(x0)))
        return bnds, x0

    def solve(self, ftol=1e-20):
        cons = self._get_cons()
        bnds, x0 = self._get_bnds()
        options = {'disp': False, 'maxiter': 500, 'ftol': ftol}

        res = minimize(self._target_func, x0, bounds=bnds, constraints=cons, method='SLSQP', options=options)
        if not res.success:
            weight = pd.Series(index=self.asset_list + ['cash'], data=[np.nan] * (len(self.asset_list) + 1))
            status = 'infeasible'
        else:
            arr_x = np.array(list(res.x))
            for i in range(len(self.asset_list)):
                if arr_x[i] < self.lb_list[i]:
                    arr_x[i] = self.lb_list[i]
            for i in range(len(self.asset_list)):
                if arr_x[i] > self.ub_list[i]:
                    arr_x[i] = self.ub_list[i]
            if sum(arr_x) != self.total_weight:
                arr_x = arr_x / arr_x.sum() * self.total_weight

            weight = pd.Series(index=self.asset_list + ['cash'], data=list(arr_x) + [1.0 - arr_x.sum()])
            status = 'optimal'
        return weight, status



# -*- coding:utf-8 -*-

"""
获取基金净值数据接口
Created on 2020/6/15
@author: Meng Lv
@group: Howbuy FDC
@contact: meng.lv@howbuy.com
"""

from __future__ import division
import time
import pandas as pd
from hbshare.db.fund import cons as ct
import hbshare as hbs
import demjson


def get_fund_newest_nav_by_code(code, retry_count=3, pause=0.01, timeout=10):
    """
        获得公募基金最新净值（含最新回报，区间回报）
    :param code:
    :return:
    """
    api = hbs.hb_api()
    for _ in range(retry_count):
        time.sleep(pause)
        ct._write_console()
        url = ct.HOWBUY_FUND_NEWEST_NAV % (ct.P_TYPE['https'], ct.DOMAINS['hbcgi'], code)

        org_js = api.query(url)
        status_code = int(org_js['common']['responseCode'])
        if status_code != 1:
            status = str(org_js['common']['responseContent'])
            raise ValueError(status)

        if 'opens' not in org_js:
            status = "未查询到该基金数据"
            raise ValueError(status)

        data = org_js['opens']
        fund_df = pd.DataFrame(data, columns=ct.HOWBUY_FUND_NEWEST_NAV_COLUMNS)
        fund_df['jjjz'] = fund_df['jjjz'].astype(float)
        fund_df['ljjz'] = fund_df['ljjz'].astype(float)
        fund_df['hbdr'] = fund_df['hbdr'].astype(float)
        fund_df['hb1y'] = fund_df['hb1y'].astype(float)
        fund_df['hb3y'] = fund_df['hb3y'].astype(float)
        fund_df['hb6y'] = fund_df['hb6y'].astype(float)
        fund_df['hbjn'] = fund_df['hbjn'].astype(float)
        fund_df['hb1n'] = fund_df['hb1n'].astype(float)
        fund_df['zfxz'] = fund_df['zfxz'].astype(float)
        return fund_df
    raise IOError(ct.NETWORK_URL_ERROR_MSG)


def get_fund_holding(fund_code, publish_date, retry_count=3, pause=0.01):
    """
        获得公募基金持仓情况
    :param fund_code: 基金代码
    :param publish_date: 公告日期
    :param retry_count: 重试次数
    :param pause: 重试间隔时间
    :return:
    """
    api = hbs.hb_api()
    for _ in range(retry_count):
        time.sleep(pause)
        ct._write_console()
        url = ct.HOWBUY_FUND_HOLDING % (ct.P_TYPE['https'], ct.DOMAINS['static'], fund_code, time.time())

        js_text = api.get(url)
        gpzh = js_text.splitlines()[6].replace(";", "")[15:]

        # 日期格式转化
        # st = time.strftime("%Y-%m-%d", time.strptime(publish_date, "%Y%m%d"))

        holding = demjson.decode(gpzh)[publish_date]

        fund_df = pd.DataFrame(holding, columns=ct.HOWBUY_FUND_HOLDING_COLUMNS)
        fund_df['zqdm'] = fund_df['zqdm'].astype(str)
        fund_df['zqmc'] = fund_df['zqmc'].astype(str)
        fund_df['zjbl'] = fund_df['zjbl'].astype(float)
        fund_df['ccdb'] = fund_df['ccdb'].astype(float)
        return fund_df
    raise IOError(ct.NETWORK_URL_ERROR_MSG)


def get_fund_holding_publish_date(fund_code, retry_count=3, pause=0.01):
    """
        获得公募基金定期报告发布时间查询
    :param fund_code: 基金代码
    :param retry_count: 重试次数
    :param pause: 重试间隔时间
    :return:
    """
    api = hbs.hb_api()
    for _ in range(retry_count):
        time.sleep(pause)
        ct._write_console()
        url = ct.HOWBUY_FUND_HOLDING % (ct.P_TYPE['https'], ct.DOMAINS['static'], fund_code, time.time())

        js_text = api.get(url)
        gpzh = js_text.splitlines()[6].replace(";", "")[15:]

        date_list = demjson.decode(gpzh)["dateList"]

        fund_df = pd.DataFrame(date_list, columns=['publish_date'])
        fund_df['publish_date'] = fund_df['publish_date'].astype(str)
        return fund_df
    raise IOError(ct.NETWORK_URL_ERROR_MSG)



"""
alpha标的超额回测模块
"""
import pandas as pd
import numpy as np
from datetime import datetime
import os
import hbshare as hbs
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_trading_day_list
from Arbitrage_backtest import cal_annual_return, cal_annual_volatility, cal_sharpe_ratio, cal_max_drawdown
from sqlalchemy import create_engine
from hbshare.quant.Kevin.rm_associated.config import engine_params, style_names, industry_names
from plotly.offline import plot as plot_ly
import statsmodels.api as sm
import plotly.graph_objs as go
import plotly.figure_factory as ff
import matplotlib.pyplot as plt
plt.style.use('seaborn')


class AlphaBacktest:
    def __init__(self, data_path, start_date, end_date, mode="ratio"):
        self.data_path = data_path
        self.start_date = start_date
        self.end_date = end_date
        self.mode = mode
        self._load_data()

    def _load_data(self):
        data_with_header = pd.read_excel(
            os.path.join(self.data_path, r"指增-{}.xlsx".format(self.end_date)), sheet_name='原始净值')
        data = pd.read_excel(
            os.path.join(self.data_path, "指增-{}.xlsx".format(self.end_date)), sheet_name='原始净值', header=1)
        data['t_date'] = data['t_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        data.index = data['t_date']
        cols = data_with_header.columns.tolist()

        trading_day_list = get_trading_day_list(self.start_date, self.end_date, frequency="week")

        data_param = []
        type_list = [x for x in cols if not x.startswith('Unnamed')]
        for i in range(len(type_list) - 1):
            if type_list[i] in ['量价（500）', '机器学习', '基本面']:
                s_index, e_index = cols.index(type_list[i]), cols.index(type_list[i + 1])
                data_slice = data[data.columns[s_index: e_index]]
                data_slice = data_slice[data_slice.index >= self.start_date].reindex(trading_day_list)
                data_param.append(data_slice)
            else:
                pass

        self.nav_data = pd.concat(data_param, axis=1).dropna(how='all', axis=0).sort_index()

        nav_df = self.nav_data.copy()
        nav_df['trade_date'] = nav_df.index
        nav_df['month'] = nav_df['trade_date'].apply(lambda x: datetime.strptime(x, '%Y%m%d').month)
        nav_df['next_month'] = nav_df['month'].shift(-1)
        nav_df.iloc[-1, -1] = nav_df['month'].tolist()[-1]
        # nav_df.iloc[-1, -1] = 7.

        self.month_list = nav_df[nav_df['month'] != nav_df['next_month']]['trade_date'].tolist()

        sql_script = "SELECT JYRQ as TRADEDATE, ZQMC as INDEXNAME, SPJG as TCLOSE from funddb.ZSJY WHERE ZQDM = '{}' " \
                     "and JYRQ >= {} and JYRQ <= {}".format('000905', self.start_date, self.end_date)
        index_data = pd.DataFrame(hbs.db_data_query('readonly', sql_script, page_size=5000)['data']).set_index('TRADEDATE')['TCLOSE']
        self.index_nav = index_data.reindex(self.nav_data.index)

        sql_script = "SELECT * FROM factor_return where trade_date >= {} and trade_date <= {}".format(
            self.start_date, self.end_date)
        engine = create_engine(engine_params)
        data = pd.read_sql(sql_script, engine)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
        factor_data = pd.pivot_table(
            data, index='trade_date', columns='factor_name', values='factor_ret').sort_index()[
            ['size', 'beta', 'momentum']]
        self.factor_data = (1 + factor_data).cumprod().reindex(self.nav_data.index)

        self.fund_date = pd.read_excel(os.path.join(self.data_path, r"fund_date.xlsx"), dtype={"start_date": str})

    def _counting_num(self):
        df = self.nav_data.copy()
        df['trade_date'] = df.index
        df['month'] = df['trade_date'].apply(lambda x: datetime.strptime(x, '%Y%m%d').month)
        df['next_month'] = df['month'].shift(-1)
        df.iloc[-1, -1] = df['month'].tolist()[-1]
        quarter_list = df[(df['month'] != df['next_month']) & (df['month'].isin([3, 6, 9, 12]))]['trade_date'].tolist()

        df = self.nav_data.copy()
        num_list = []
        for i in range(len(quarter_list) - 2):
            start_date, end_date = quarter_list[i], quarter_list[i + 2]
            period_data = df.loc[start_date: end_date, :].dropna(axis=1)
            num_list.append(period_data.shape[1])

        count_df = pd.DataFrame(index=quarter_list[1: -1], data=num_list, columns=['number'])

        return count_df

    @staticmethod
    def _regression(y, x):
        return sm.OLS(y, sm.add_constant(x)).fit().resid

    @staticmethod
    def _single_period_calc(ret, factor_ret, rf=0.015):
        annual_return = ret.apply(cal_annual_return, axis=0)
        annual_vol = ret.apply(cal_annual_volatility, axis=0)
        sharpe_ratio = (annual_return - rf) / annual_vol

        f_ret = sm.add_constant(np.array(factor_ret))
        resid = ret.apply(lambda x: sm.OLS(x, f_ret).fit().resid)
        alpha = (1 + resid).prod() - 1

        max_drawdown = (1 + ret).cumprod().apply(cal_max_drawdown, axis=0)

        factor = annual_return.to_frame('ret').merge(
            sharpe_ratio.to_frame('sharpe'), left_index=True, right_index=True).merge(
            alpha.to_frame('s_ret'), left_index=True, right_index=True).merge(
            max_drawdown.to_frame('max_drawdown'), left_index=True, right_index=True)

        return factor

    @staticmethod
    def _calc_group_weight(factor_series, group_num=5):
        n = factor_series.shape[0]
        p_list = [round((n / group_num) * i, 1) for i in range(0, group_num + 1)]
        group_df = pd.DataFrame(index=factor_series.index, columns=['group_{}'.format(i) for i in range(1, group_num + 1)])
        group_df = pd.merge(
            factor_series.to_frame('factor'), group_df, left_index=True, right_index=True).sort_values(by='factor')

        for i in range(len(p_list) - 1):
            point1, point2 = p_list[i], p_list[i + 1]
            tmp1 = group_df.index[np.floor(point1).astype(int)]
            if i == len(p_list) - 2:
                tmp2 = group_df.index[np.floor(point2).astype(int) - 1]
            else:
                tmp2 = group_df.index[np.floor(point2).astype(int)]
            group_df.loc[tmp1: tmp2, 'group_{}'.format(i + 1)] = 1.

        for i in range(1, len(p_list) - 1):
            point = p_list[i]
            tmp = group_df.index[np.floor(point).astype(int)]
            t_values = round(point - int(point), 2)
            group_df.loc[tmp, 'group_{}'.format(i)] = t_values
            group_df.loc[tmp, 'group_{}'.format(i + 1)] = round(1 - t_values, 1)

        group_df = group_df.fillna(0.)

        return group_df

    def calculate_factor(self, start_date, calc_period, hold_period):
        """
        start_date: 有效数据的起始时间点
        calc_period: 回溯区间长度
        hold_period: 持有区间长度
        """
        nav_df = self.nav_data.copy()
        reb_list = [x for x in self.month_list if x >= start_date][::hold_period]

        # 实际运行起始
        tmp = self.fund_date.set_index('fund_name')
        for name in tmp.index.tolist():
            e_date = tmp.loc[name, "start_date"]
            if nav_df.index.tolist().index(e_date) != 0:
                t_index = nav_df.index.tolist().index(e_date) - 1
                nav_df.loc[: nav_df.index.tolist()[t_index], name] = np.NAN
            else:
                pass

        # 异常值剔除
        include_list = [x for x in nav_df.columns if x not in [
            '聚宽一号', '无量1期', '思勰投资-中证500指数增强1号', '希格斯水手2号', '龙旗红旭500指数增强']]
        nav_df = nav_df[include_list]

        factor_list = []
        group_ret_list = []
        group_ret_list2 = []
        corr_list = []
        group_num = 5
        for reb_date in reb_list:
            s_date = self.month_list[self.month_list.index(reb_date) - calc_period]
            if self.month_list.index(reb_date) + hold_period >= len(self.month_list):
                e_date = self.month_list[-1]
            else:
                e_date = self.month_list[self.month_list.index(reb_date) + hold_period]
            # data check
            period_data = nav_df.loc[s_date: e_date].dropna(axis=1)

            calc_data = period_data.loc[s_date: reb_date]
            bm_data = self.index_nav.loc[s_date: reb_date]
            assert (calc_data.shape[0] == bm_data.shape[0])
            excess_ret = calc_data.pct_change().dropna().sub(bm_data.pct_change().dropna().squeeze(), axis=0)
            factor_ret = self.factor_data.pct_change().dropna().reindex(excess_ret.index)

            factor_ret = factor_ret[['size']]

            factor = self._single_period_calc(excess_ret, factor_ret)
            factor['fund_name'] = factor.index
            factor['reb_date'] = reb_date
            factor.reset_index(drop=True, inplace=True)
            # factor1 = factor.copy()

            # group weight
            # factor['calmar'] = factor['ret'] / factor['max_drawdown'].abs()
            factor_series = factor.set_index('fund_name')['s_ret']
            weight_df = self._calc_group_weight(factor_series, group_num)
            print(reb_date, weight_df.shape[0])

            # holding ret
            future_data = period_data.loc[reb_date: e_date]
            bm_data_fu = self.index_nav.loc[reb_date: e_date]
            assert (future_data.shape[0] == bm_data_fu.shape[0])
            excess_fu = future_data.pct_change().fillna(0.).sub(bm_data_fu.pct_change().fillna(0.).squeeze(), axis=0)
            excess_df = (1 + excess_fu).cumprod().T.reindex(weight_df.index).T
            group_nav = excess_df.dot(weight_df[weight_df.columns[1:]])
            group_nav = group_nav / group_nav.iloc[0]
            group_ret = group_nav.pct_change().dropna()

            a = ((1 + excess_fu).prod() - 1).to_frame('excess')
            a = pd.merge(factor, a, left_on='fund_name', right_index=True)
            corr_list.append(a['ret'].corr(a['excess']))
            # print(reb_date, a['ret'].corr(a['excess']))

            group_ret_list.append(group_ret)
            group_ret_list2.append(((1 + group_ret).prod() - 1).to_frame(reb_date).T)
            factor_list.append(factor)

        factor_df = pd.concat(factor_list)
        group_ret_all = pd.concat(group_ret_list)
        (1 + group_ret_all).cumprod().plot.line()

        # annual_alpha = group_ret_all.apply(cal_annual_return, axis=0)
        year_alpha = pd.DataFrame(
            index=['2019', '2020', '2021'], columns=['group_{}'.format(i) for i in range(1, group_num + 1)])
        year_alpha.loc['2019'] = (1 + group_ret_all[:"20191227"]).prod() - 1
        year_alpha.loc['2020'] = (1 + group_ret_all["20200103": "20201231"]).prod() - 1
        year_alpha.loc['2021'] = (1 + group_ret_all["20210108":]).prod() - 1

        return factor_df, group_ret_all

    def run(self, start_date):
        nav_df = self.nav_data.copy()
        reb_list = [x for x in self.month_list if x >= start_date][::3]

        # 实际运行起始
        tmp = self.fund_date.set_index('fund_name')
        for name in tmp.index.tolist():
            e_date = tmp.loc[name, "start_date"]
            if nav_df.index.tolist().index(e_date) != 0:
                t_index = nav_df.index.tolist().index(e_date) - 1
                nav_df.loc[: nav_df.index.tolist()[t_index], name] = np.NAN
            else:
                pass

        # 异常值剔除
        include_list = [x for x in nav_df.columns if x not in [
            '聚宽一号', '无量1期', '思勰投资-中证500指数增强1号', '希格斯水手2号', '龙旗红旭500指数增强']]
        nav_df = nav_df[include_list]

        factor_list = []
        group_ret_list = []
        group_num = 4
        weight_list = []
        for reb_date in reb_list:
            s_date = self.month_list[self.month_list.index(reb_date) - 6]
            pre_date = self.month_list[self.month_list.index(reb_date) - 3]
            if self.month_list.index(reb_date) + 3 >= len(self.month_list):
                e_date = self.month_list[-1]
            else:
                e_date = self.month_list[self.month_list.index(reb_date) + 3]
            # data check
            period_data = nav_df.loc[s_date: e_date].dropna(axis=1)

            calc_data = period_data.loc[s_date: reb_date]
            bm_data = self.index_nav.loc[s_date: reb_date]
            assert (calc_data.shape[0] == bm_data.shape[0])
            excess_ret = calc_data.pct_change().dropna().sub(bm_data.pct_change().dropna().squeeze(), axis=0)
            factor_ret = self.factor_data.pct_change().dropna().reindex(excess_ret.index)

            factor_ret = factor_ret[['size']]
            # 6个月因子
            factor = self._single_period_calc(excess_ret, factor_ret)
            factor['fund_name'] = factor.index
            factor['reb_date'] = reb_date
            factor.reset_index(drop=True, inplace=True)
            factor1 = factor.copy()
            # 3个月因子
            factor = self._single_period_calc(excess_ret.loc[pre_date:][1:], factor_ret.loc[pre_date:][1:])
            factor['fund_name'] = factor.index
            factor['reb_date'] = reb_date
            factor.reset_index(drop=True, inplace=True)
            factor2 = factor.copy()
            factor2.rename(columns={"ret": "ret_3M"}, inplace=True)

            factor = pd.merge(factor1, factor2[['fund_name', 'ret_3M']], on='fund_name')
            list1 = factor[factor['max_drawdown'] >= factor['max_drawdown'].quantile(0.25)]['fund_name'].tolist()
            tmp = factor[factor['fund_name'].isin(list1)]
            selected_name = tmp[tmp['sharpe'] >= tmp['sharpe'].quantile(0.66)]['fund_name'].tolist()

            # group weight
            factor_series = factor.set_index('fund_name')['ret_3M']
            weight_df = self._calc_group_weight(factor_series, group_num)
            print(reb_date, weight_df.shape[0], len(selected_name))

            weight_df.loc[selected_name, 'port'] = 1.0
            weight_df['port'] = weight_df['port'].fillna(0.)

            tmp = weight_df.copy()
            tmp['date'] = reb_date
            weight_list.append(tmp[tmp['port'] == 1][['date', 'port']])
            # weight_list.append(tmp[tmp['group_5'] > 1e-6][['date', 'group_5']])

            # holding ret
            future_data = period_data.loc[reb_date: e_date]
            bm_data_fu = self.index_nav.loc[reb_date: e_date]
            assert (future_data.shape[0] == bm_data_fu.shape[0])
            excess_fu = future_data.pct_change().fillna(0.).sub(bm_data_fu.pct_change().fillna(0.).squeeze(), axis=0)
            excess_df = (1 + excess_fu).cumprod().T.reindex(weight_df.index).T
            group_nav = excess_df.dot(weight_df[weight_df.columns[1:]])
            group_nav = group_nav / group_nav.iloc[0]
            group_ret = group_nav.pct_change().dropna()

            # a = ((1 + excess_fu).prod() - 1).to_frame('excess')
            # a = pd.merge(factor, a, left_on='fund_name', right_index=True)
            # print(reb_date, a['ret'].corr(a['excess']))

            group_ret_list.append(group_ret)
            factor_list.append(factor)

        factor_df = pd.concat(factor_list)
        group_ret_all = pd.concat(group_ret_list)
        (1 + group_ret_all).cumprod().plot.line()

        annual_alpha = group_ret_all.apply(cal_annual_return, axis=0)
        year_alpha = pd.DataFrame(
            index=['2019', '2020', '2021'], columns=['group_{}'.format(i) for i in range(1, group_num + 1)] + ['port'])
        year_alpha.loc['2019'] = (1 + group_ret_all[:"20191227"]).prod() - 1
        year_alpha.loc['2020'] = (1 + group_ret_all["20200103": "20201231"]).prod() - 1
        year_alpha.loc['2021'] = (1 + group_ret_all["20210108":]).prod() - 1

        weight_all = pd.concat(weight_list).reset_index()
        weight_all['weight'] = 1.
        weight_table = pd.pivot_table(weight_all, index='date', columns='fund_name', values='weight').sort_index().T
        weight_table[weight_table.columns[4:]].fillna(0.).to_clipboard()

        # 添加申购T+5的回测
        group_ret_list_adjusted = []
        next_weight = 1.
        for reb_date in reb_list:

            t_weight = weight_table[reb_date].dropna()
            t_weight /= t_weight.sum()

            if self.month_list.index(reb_date) + 3 >= len(self.month_list):
                e_date = self.month_list[-1]
            else:
                e_date = self.month_list[self.month_list.index(reb_date) + 3]
            future_data = nav_df.loc[reb_date: e_date].dropna(axis=1)
            bm_data_fu = self.index_nav.loc[reb_date: e_date]

            assert (future_data.shape[0] == bm_data_fu.shape[0])

            excess_fu = future_data.pct_change().fillna(0.).sub(bm_data_fu.pct_change().fillna(0.).squeeze(), axis=0)

            if reb_date == reb_list[0]:
                excess_df = (1 + excess_fu).cumprod().T.reindex(t_weight.index).T
                group_nav = excess_df.dot(t_weight)
                group_ret = group_nav.pct_change().dropna()
                next_weight = excess_df.multiply(t_weight).iloc[-1]
                next_weight /= next_weight.sum()
                group_ret_list_adjusted.append(group_ret)
            if reb_date != reb_list[0]:
                weight_compare = next_weight.to_frame('former').merge(
                    t_weight.to_frame('latter'), left_index=True, right_index=True, how='right').fillna(0.)
                weight_compare['delta'] = weight_compare['latter'] - weight_compare['former']
                weight_compare['delta'] = weight_compare['delta'].apply(lambda x: max(x, 0))
                weight_compare['actual'] = np.where(
                    weight_compare['delta'] > 0, weight_compare['former'], weight_compare['latter'])
                excess_df1 = (1 + excess_fu).cumprod().T.reindex(weight_compare.index).T
                tmp = excess_fu[1:].copy()
                tmp.iloc[0] = 0.
                excess_df2 = (1 + tmp).cumprod().T.reindex(weight_compare.index).T

                w_l = excess_df1.multiply(weight_compare['actual']) + excess_df2.multiply(weight_compare['delta']).reindex(
                    excess_df1.index).fillna(method='bfill')
                group_nav = w_l.sum(axis=1)
                group_ret = group_nav.pct_change().dropna()
                next_weight = w_l.iloc[-1]
                next_weight /= next_weight.sum()
                group_ret_list_adjusted.append(group_ret)

        group_ret_adjusted = pd.concat(group_ret_list_adjusted)
        compare_df = (1 + group_ret_all['port']).cumprod().to_frame('normal').merge(
            (1 + group_ret_adjusted).cumprod().to_frame('adjusted'), left_index=True, right_index=True)

        return factor_df, group_ret_all


if __name__ == '__main__':
    # AlphaBacktest("D:\\量化产品跟踪\\指数增强", '20161230', '20220520').run()
    # f_df, g_ret = AlphaBacktest("D:\\量化产品跟踪\\指数增强", '20161230', '20220520').calculate_factor('20181228', 3, 3)
    AlphaBacktest("D:\\量化产品跟踪\\指数增强", '20161230', '20220708').run('20181228')


# -*- coding: utf-8 -*-

from hbshare.fe.asset_allocation.data_loader import Loader
from hbshare.fe.common.util.exception import InputParameterError
from hbshare.fe.common.util.logger import logger
from hbshare.fe.common.util.verifier import verify_type
import numpy as np
import pandas as pd


class Simple:
    def __init__(self, asset_type, asset_list, method, start_date, end_date, is_reblance, reblance_type, n, frequency, set_w=None):
        self.asset_type = asset_type
        self.asset_list = asset_list
        self.method = method
        self.start_date = start_date
        self.end_date = end_date
        self.is_reblance = is_reblance
        self.reblance_type = reblance_type
        self.n = n
        self.frequency = frequency
        self.set_w = set_w if set_w is not None else [1.0 / len(self.asset_list)] * len(self.asset_list)
        self._verify_input_param()
        self._load()
        self._init_data()

    def _verify_input_param(self):
        verify_type(self.asset_type, 'asset_type', str)
        verify_type(self.asset_list, 'asset_list', list)
        verify_type(self.method, 'method', str)
        verify_type(self.start_date, 'start_date', str)
        verify_type(self.end_date, 'end_date', str)
        verify_type(self.is_reblance, 'is_reblance', bool)
        verify_type(self.reblance_type, 'reblance_type', str)
        verify_type(self.n, 'n', int)
        verify_type(self.frequency, 'frequency', str)
        if self.asset_type not in ['index', 'fund']:
            msg = "asset_type not supported, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)
        if len(self.asset_list) == 0:
            msg = "asset_list is empty, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)
        if self.method not in ['simple']:
            msg = "method not supported, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)
        if self.reblance_type not in ['init_weight', 'init_target']:
            msg = "reblance_type not supported, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)
        if self.n <= 0:
            msg = "n must be larger than 0, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)
        if self.frequency not in ['day', 'week', 'month', 'quarter', 'year']:
            msg = "frequency not supported, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)
        if len(self.set_w) != len(self.asset_list):
            msg = "set_w is not the same length with asset_list, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)
        if len([i for i in self.set_w if (i >= 0.0 and i <= 1.0)]) != len(self.asset_list):
            msg = "set_w must be between 0 and 1, check your input"
            logger.error(msg)
            raise InputParameterError(message=msg)

    def _load(self):
        self.calendar_df = Loader().read_cal('19900101', self.end_date)
        self.calendar_df = self.calendar_df.rename(columns={'JYRQ': 'CALENDAR_DATE', 'SFJJ': 'IS_OPEN', 'SFZM': 'IS_WEEK_END', 'SFYM': 'IS_MONTH_END'})
        self.calendar_df['CALENDAR_DATE'] = self.calendar_df['CALENDAR_DATE'].astype(str)
        self.calendar_df = self.calendar_df.sort_values('CALENDAR_DATE')
        self.calendar_df['IS_OPEN'] = self.calendar_df['IS_OPEN'].astype(int).replace({0: 1, 1: 0})
        self.calendar_df['IS_WEEK_END'] = self.calendar_df['IS_WEEK_END'].fillna(0).astype(int)
        self.calendar_df['IS_MONTH_END'] = self.calendar_df['IS_MONTH_END'].fillna(0).astype(int)
        self.calendar_df['YEAR_MONTH'] = self.calendar_df['CALENDAR_DATE'].apply(lambda x: x[:6])
        self.calendar_df['MONTH'] = self.calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:6])
        self.calendar_df['MONTH_DAY'] = self.calendar_df['CALENDAR_DATE'].apply(lambda x: x[4:])
        self.calendar_df['IS_QUARTER_END'] = np.where((self.calendar_df['IS_MONTH_END'] == 1) & (self.calendar_df['MONTH'].isin(['03', '06', '09', '12'])), 1, 0)
        self.calendar_df['IS_QUARTER_END'] = self.calendar_df['IS_QUARTER_END'].astype(int)
        self.calendar_df['IS_SEASON_END'] = np.where(self.calendar_df['MONTH_DAY'].isin(['0331', '0630', '0930', '1231']), 1, 0)
        self.calendar_df['IS_SEASON_END'] = self.calendar_df['IS_SEASON_END'].astype(int)
        self.trade_cal = self.calendar_df[self.calendar_df['IS_OPEN'] == 1]
        self.trade_cal['TRADE_DATE'] = self.trade_cal['CALENDAR_DATE']
        self.calendar_df = self.calendar_df.merge(self.trade_cal[['CALENDAR_DATE', 'TRADE_DATE']], on=['CALENDAR_DATE'], how='left')
        self.calendar_df['TRADE_DATE'] = self.calendar_df['TRADE_DATE'].fillna(method='ffill')
        self.calendar_df = self.calendar_df[['CALENDAR_DATE', 'IS_OPEN', 'IS_WEEK_END', 'IS_MONTH_END', 'IS_QUARTER_END', 'IS_SEASON_END', 'TRADE_DATE', 'YEAR_MONTH', 'MONTH', 'MONTH_DAY']]
        return

    def _init_data(self):
        # 确定再平衡时点
        self.start_trade_date = self.calendar_df[self.calendar_df['CALENDAR_DATE'] == self.start_date]['TRADE_DATE'].values[0]
        if not self.is_reblance:
            reblance_list = [self.start_trade_date]
        else:
            if self.frequency == 'day':
                reblance_list = self.calendar_df[self.calendar_df['IS_OPEN'] == 1]['TRADE_DATE'].unique().tolist()
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date) and (date <= self.end_date)]
                reblance_list = reblance_list[::self.n]
            elif self.frequency == 'week':
                reblance_list = self.calendar_df[self.calendar_df['IS_OPEN'] == 1]['TRADE_DATE'].unique().tolist()
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date) and (date <= self.end_date)]
                reblance_list = reblance_list[::self.n * 5]
            elif self.frequency == 'month':
                reblance_list = self.calendar_df['YEAR_MONTH'].unique().tolist()
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date[:6]) and (date <= self.end_date[:6])]
                reblance_list = reblance_list[::self.n]
                reblance_list = [date + self.start_trade_date[6:] for date in reblance_list]
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date) and (date <= self.end_date)]
                reblance_list = [self.calendar_df[(self.calendar_df['IS_OPEN'] == 1) & (self.calendar_df['CALENDAR_DATE'] <= date)]['TRADE_DATE'].iloc[-1] for date in reblance_list]
            elif self.frequency == 'quarter':
                reblance_list = self.calendar_df['YEAR_MONTH'].unique().tolist()
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date[:6]) and (date <= self.end_date[:6])]
                reblance_list = reblance_list[::self.n * 3]
                reblance_list = [date + self.start_trade_date[6:] for date in reblance_list]
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date) and (date <= self.end_date)]
                reblance_list = [self.calendar_df[(self.calendar_df['IS_OPEN'] == 1) & (self.calendar_df['CALENDAR_DATE'] <= date)]['TRADE_DATE'].iloc[-1] for date in reblance_list]
            elif self.frequency == 'year':
                reblance_list = self.calendar_df['YEAR_MONTH'].unique().tolist()
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date[:6]) and (date <= self.end_date[:6])]
                reblance_list = reblance_list[::self.n * 12]
                reblance_list = [date + self.start_trade_date[6:] for date in reblance_list]
                reblance_list = [date for date in reblance_list if (date >= self.start_trade_date) and (date <= self.end_date)]
                reblance_list = [self.calendar_df[(self.calendar_df['IS_OPEN'] == 1) & (self.calendar_df['CALENDAR_DATE'] <= date)]['TRADE_DATE'].iloc[-1] for date in reblance_list]
            else:
                reblance_list = [self.start_trade_date]
        self.reblance_list = reblance_list
        return

    def get_all(self):
        weight_list, status_list = [], []
        for date in self.reblance_list:
            weight = pd.DataFrame(index=[date], columns=self.asset_list)
            weight.loc[date] = self.set_w
            status = pd.DataFrame(index=[date], columns=['优化状态'], data=['optimal'])
            weight_list.append(weight)
            status_list.append(status)
            print('[{0}]/[{1}] done'.format(date, self.method))
        weight_df = pd.concat(weight_list)
        status_df = pd.concat(status_list)
        if self.reblance_type == 'init_weight':
            weight_df.iloc[1:] = weight_df.iloc[0]
            status_df.iloc[1:] = status_df.iloc[0]
        return weight_df, status_df


if __name__ == '__main__':
    # mutual_index: ['HM0001', 'HM0024', 'HM0095']
    # private_index: ['HB0000', 'HB0016', 'HB1001']
    # market_index: ['000300', '000906', 'CBA00301']
    # mutual_fund: ['002943', '688888', '000729']
    # private_fund: ['SGK768', 'SX8958', 'SR4480']

    weight_df, status_df = Simple(asset_type='index',  # index, fund
                                  asset_list=['000300', '000906', 'CBA00301'],
                                  method='simple',  # simple
                                  start_date='20181231',
                                  end_date='20220704',
                                  is_reblance=True,   # True, False
                                  reblance_type='init_target',  # init_weight，init_target
                                  n=3,
                                  frequency='month',  # day, week, month, quarter, year
                                  set_w=[0.3, 0.3, 0.4]
                                  ).get_all()
    print(weight_df, status_df)


"""
S&P Access China Analysis
"""
import pandas as pd
import hbshare as hbs
import os
from hbshare.fe.common.util.config import style_name, industry_name
from hbshare.fe.common.util.config import factor_map_dict
from plotly.offline import plot as plot_ly
import plotly.graph_objs as go
from WindPy import w

w.start()


class AccessChinaAnalysis:
    def __init__(self, trade_date, benchmark_id):
        self.trade_date = trade_date
        self.benchmark_id = benchmark_id
        self._load_data()

    def _load_portfolio_weight_series(self):
        # data set
        res = w.wset("sectorconstituent", "date={};sectorid=1000025141000000".format(self.trade_date))
        if res.ErrorCode != 0:
            data_set = pd.DataFrame()
            print("Fetching sectorconstituent data error!")
        else:
            data_set = pd.DataFrame(res.Data, index=res.Fields, columns=res.Codes).T
        # use wsd for quota exceed
        code_list = data_set['wind_code'].tolist()
        res = w.wsd(','.join(code_list), "mkt_freeshares", self.trade_date, self.trade_date, "unit=1")
        if res.ErrorCode != 0:
            data = pd.DataFrame()
            print("Fetching mkt_freeshares data error!")
        else:
            data = pd.DataFrame(res.Data, index=res.Fields, columns=res.Codes).T

        data = pd.merge(data_set, data, left_on='wind_code', right_index=True)
        data['ticker'] = data['wind_code'].apply(lambda x: x.split('.')[0])
        data['weight'] = data['MKT_FREESHARES'] / data['MKT_FREESHARES'].sum()

        return data.set_index('ticker')['weight']

    def _load_benchmark_weight_series(self, ticker=None):
        if ticker is None:
            ticker = self.benchmark_id
        sql_script = "SELECT * FROM hsjy_gg.SecuMain where SecuCategory = 4 and SecuCode = '{}'".format(
            ticker)
        res = hbs.db_data_query('readonly', sql_script)
        index_info = pd.DataFrame(res['data'])
        inner_code = index_info.set_index('SECUCODE').loc[ticker, 'INNERCODE']

        sql_script = "SELECT (select a.SecuCode from hsjy_gg.SecuMain a where a.InnerCode = b.InnerCode and " \
                     "rownum = 1) SecuCode, b.EndDate, b.Weight FROM hsjy_gg.LC_IndexComponentsWeight b WHERE " \
                     "b.IndexCode = '{}' and b.EndDate = to_date('{}', 'yyyymmdd')".format(inner_code, self.trade_date)
        data = pd.DataFrame(hbs.db_data_query('readonly', sql_script)['data'])
        weight_df = data.rename(
            columns={"SECUCODE": "consTickerSymbol", "ENDDATE": "effDate", "WEIGHT": "weight"})

        return weight_df.set_index('consTickerSymbol')['weight'] / 100.

    def _load_style_exposure(self):
        sql_script = "SELECT * FROM st_ashare.r_st_barra_style_factor where TRADE_DATE = '{}'".format(self.trade_date)
        res = hbs.db_data_query('alluser', sql_script, page_size=5000)
        exposure_df = pd.DataFrame(res['data']).set_index('ticker')
        ind_names = [x.lower() for x in industry_name['sw'].values()]
        exposure_df = exposure_df[style_name + ind_names]

        return exposure_df

    def _load_data(self):
        portfolio_weight_series = self._load_portfolio_weight_series()
        benchmark_weight_series = self._load_benchmark_weight_series()
        # 300 & 500
        benchmark_weight_300 = self._load_benchmark_weight_series('000300')
        benchmark_weight_500 = self._load_benchmark_weight_series('000905')
        style_exposure_df = self._load_style_exposure()

        self.data_param = {
            "portfolio_weight_series": portfolio_weight_series,
            "benchmark_weight_series": benchmark_weight_series,
            "benchmark_weight_300": benchmark_weight_300,
            "benchmark_weight_500": benchmark_weight_500,
            "style_exposure_df": style_exposure_df
        }

    @staticmethod
    def plotly_style_bar(df, title_text, figsize=(1200, 800), legend_x=0.30):
        fig_width, fig_height = figsize
        cols = df.columns.tolist()
        color_list = ['rgb(49, 130, 189)', 'rgb(204, 204, 204)', 'rgb(216, 0, 18)']
        data = []
        for i in range(len(cols)):
            col = cols[i]
            trace = go.Bar(
                x=df.index.tolist(),
                y=df[col],
                name=col,
                marker=dict(color=color_list[i])
            )
            data.append(trace)

        layout = go.Layout(
            title=dict(text=title_text),
            autosize=False, width=fig_width, height=fig_height,
            yaxis=dict(tickfont=dict(size=12), showgrid=True),
            xaxis=dict(showgrid=True),
            legend=dict(orientation="h", x=legend_x),
            template='plotly_white'
        )

        fig = go.Figure(data=data, layout=layout)

        return fig

    def get_construct_result(self):
        portfolio_weight_series = self.data_param.get('portfolio_weight_series')
        benchmark_weight_series = self.data_param.get('benchmark_weight_series')
        style_exposure_df = self.data_param.get('style_exposure_df')
        # benchmark_weight_300 = self.data_param.get('benchmark_weight_300')
        # benchmark_weight_500 = self.data_param.get('benchmark_weight_500')

        # 板块分布
        weight_df = portfolio_weight_series.reset_index().rename(columns={"wind_code": "ticker"})
        weight_df.loc[weight_df['ticker'].str.startswith('0'), 'sector'] = '深市'
        weight_df.loc[weight_df['ticker'].str.startswith('60'), 'sector'] = '沪市'
        weight_df.loc[weight_df['ticker'].str.startswith('30'), 'sector'] = '创业板'
        weight_df.loc[weight_df['ticker'].str.startswith('688'), 'sector'] = '科创板'
        sector_distribution = weight_df.groupby('sector')['weight'].sum()

        # 风格及行业分布
        idx = portfolio_weight_series.index.union(benchmark_weight_series.index).intersection(
            style_exposure_df.index)

        portfolio_weight_series = portfolio_weight_series.reindex(idx).fillna(0.)
        benchmark_weight_series = benchmark_weight_series.reindex(idx).fillna(0.)
        style_exposure_df = style_exposure_df.reindex(idx).astype(float)
        portfolio_expo = style_exposure_df.T.dot(portfolio_weight_series)
        benchmark_expo = style_exposure_df.T.dot(benchmark_weight_series)
        style_expo = pd.concat([portfolio_expo.to_frame('port'), benchmark_expo.to_frame('bm')], axis=1)
        style_expo['active'] = style_expo['port'] - style_expo['bm']

        reverse_ind = dict([(value.lower(), key) for (key, value) in industry_name['sw'].items()])
        benchmark_id_map = {"000300": "沪深300", "000905": "中证500", "000906": "中证800", "000852": "中证1000"}

        if not os.path.exists('D:\\市场微观结构图\\S&P_Access_China\\{}'.format(self.trade_date)):
            os.mkdir('D:\\市场微观结构图\\S&P_Access_China\\{}'.format(self.trade_date))

        style_df = style_expo[['port', 'bm', 'active']].rename(
            columns={"port": "S&P Access China", "bm": benchmark_id_map[self.benchmark_id], "active": "主动暴露"}).loc[
            style_name]
        style_df.index = style_df.index.map(factor_map_dict)
        fig = self.plotly_style_bar(style_df, "横截面持仓风格暴露")
        plot_ly(fig, filename="D:\\市场微观结构图\\S&P_Access_China\\{}\\风格暴露.html".format(self.trade_date))

        # 行业
        ind_df = style_expo[['port', 'bm', 'active']].rename(
            columns={"port": "S&P Access China", "bm": benchmark_id_map[self.benchmark_id],
                     "active": "主动暴露"}).iloc[10:]
        ind_df.index = [reverse_ind[x] for x in ind_df.index]
        fig = self.plotly_style_bar(ind_df, "横截面持仓行业暴露", figsize=(1500, 700), legend_x=0.35)
        plot_ly(fig, filename="D:\\市场微观结构图\\S&P_Access_China\\{}\\行业暴露.html".format(self.trade_date))

        results = {"style_df": style_df, "ind_df": ind_df, "sector_distribution": sector_distribution}

        return results


if __name__ == '__main__':
    AccessChinaAnalysis('20210831', '000906').get_construct_result()